# coding=utf-8
"""Communities module."""
from __future__ import absolute_import

from abilian.sbe.extension import sbe


def register_plugin(app):
    sbe.init_app(app)

    # Used for side-effect
    from . import events  # noqa

    from .views import communities

    app.register_blueprint(communities)

    from . import search

    search.init_app(app)



# coding=utf-8
"""Set up the markdown converter.

Add extensions here (for now).
"""
from __future__ import absolute_import, print_function, unicode_literals

import markdown
from flask import url_for
from markdown.extensions.wikilinks import WikiLinkExtension, WikiLinks
from markdown.util import etree

__all__ = ("convert", "SBEWikiLinkExtension")


class UrlBuilder(object):
    def __init__(self, page):
        self.page = page

    def build(self, label, base, end):
        label = label.strip()
        return url_for(".page", community_id=self.page.community.slug, title=label)


def convert(page, text):
    build_url = UrlBuilder(page).build
    extension = SBEWikiLinkExtension([("build_url", build_url)])
    ctx = {}
    ctx["extensions"] = [extension, "toc"]
    ctx["output_format"] = "html5"
    md = markdown.Markdown(**ctx)
    return md.convert(text)


class SBEWikiLinkExtension(WikiLinkExtension):
    def extendMarkdown(self, md, md_globals):
        self.md = md

        # append to end of inline patterns
        WIKILINK_RE = r"\[\[(.*?)\]\]"
        wikilinkPattern = SBEWikiLinks(WIKILINK_RE, self.getConfigs())
        wikilinkPattern.md = md
        md.inlinePatterns.add("wikilink", wikilinkPattern, "<not_strong")


class SBEWikiLinks(WikiLinks):
    def handleMatch(self, m):
        from .forms import page_exists

        if m.group(2).strip():
            base_url, end_url, html_class = self._getMeta()
            label = m.group(2).strip()
            url = self.config["build_url"](label, base_url, end_url)
            a = etree.Element("a")
            a.text = label
            a.set("href", url)
            if html_class:
                if page_exists(label):
                    a.set("class", html_class)
                else:
                    a.set("class", html_class + " new")
        else:
            a = ""
        return a



# coding=utf-8
# Note: this test suite is using pytest instead of the unittest-based scaffolding
# provided by SBE. Hopefully one day all of SBE will follow.

from __future__ import absolute_import, print_function, unicode_literals

from datetime import datetime, timedelta

import abilian.i18n
import pytest
from abilian.core.signals import activity

from abilian.sbe.app import create_app
from abilian.sbe.apps.communities.common import activity_time_format


@pytest.fixture
def app():
    app = create_app()

    # We need some incantations here to make babel work in the test
    babel = abilian.i18n.babel
    babel.locale_selector_func = None

    yield app

    # Signals are globals and apparently need to be cleaned up.
    # At this point, only the "activity" signal seems to have a side effect.
    activity._clear_state()


def test_activity_time_format(app):
    # We need the app context because of Babel.

    with app.app_context():
        then = datetime(2017, 1, 1, 12, 0, 0)

        now = then + timedelta(0, 5)
        assert activity_time_format(then, now) == "5s"

        now = then + timedelta(0, 5 * 60)
        assert activity_time_format(then, now) == "5m"

        now = then + timedelta(0, 5 * 60 * 60)
        assert activity_time_format(then, now) == "5h"

        now = then + timedelta(1, 5)
        assert activity_time_format(then, now) == "1d"

        now = then + timedelta(60, 5)
        assert activity_time_format(then, now) == "Jan 1"

        now = then + timedelta(365 + 60, 5)
        assert activity_time_format(then, now) == "Jan 2017"



# coding=utf-8
from __future__ import absolute_import, print_function, unicode_literals

from abilian.testing import BaseTestCase as BaseBaseTestCase
from abilian.testing import TestConfig

from .app import Application


class ConfigForTests(TestConfig):
    CELERY_ALWAYS_EAGER = True  # run tasks locally, no async
    CELERY_EAGER_PROPAGATES_EXCEPTIONS = True


class BaseTestCase(BaseBaseTestCase):
    application_class = Application
    config_class = ConfigForTests

    def get_setup_config(self):
        """Called before creating application class."""
        config = BaseBaseTestCase.get_setup_config(self)
        if hasattr(self, "no_login"):
            config.NO_LOGIN = self.no_login

        return config



# coding=utf-8
from __future__ import absolute_import, print_function, unicode_literals

from abilian.web import url_for

from abilian.sbe.testing import BaseTestCase


class TestViews(BaseTestCase):
    no_login = True

    def test_all_registered_urls(self):

        SKIP = frozenset(
            [
                "admin.audit_search_users",
                "search.search_main",
                "search.live",
                "notifications.debug_social",
                "social.groups_json",
                "social.groups_new",
                "social.users_json",
                "social.users_dt_json",
                "communities.community_default_image",
                "images.user_default",
            ]
        )

        rules = sorted(self.app.url_map.iter_rules(), key=lambda x: x.endpoint)
        for rule in rules:

            if "GET" not in rule.methods:
                continue

            if rule.arguments:
                continue

            # FIXME. Skip several exceptions.
            if rule.endpoint in SKIP:
                continue

            # These endpoints expect a parameter ('q') that we don't provide
            if rule.endpoint.endswith("list_json2"):
                continue

            url = url_for(rule.endpoint)

            if "/ajax/" in url:
                continue

            print(rule.endpoint, url)

            try:
                response = self.client.get(url)
                err_msg = "Bad link: {} (status={})".format(url, response.status_code)
                assert response.status_code in (200, 302), err_msg
            except BaseException:
                print("Problem with url: {}".format(url))
                raise



# coding=utf-8
""""""
from __future__ import absolute_import, print_function, unicode_literals

import bleach
from abilian.i18n import _l
from abilian.web.forms import Form
from abilian.web.forms.fields import DateTimeField
from abilian.web.forms.filters import strip
from abilian.web.forms.validators import required
from abilian.web.forms.widgets import RichTextWidget
from wtforms import StringField, TextAreaField, ValidationError
from wtforms.fields.html5 import URLField

ALLOWED_TAGS = [
    "a",
    "abbr",
    "acronym",
    "b",
    "blockquote",
    "br",
    "code",
    "em",
    "i",
    "li",
    "ol",
    "strong",
    "ul",
    "h1",
    "h2",
    "h3",
    "h4",
    "h5",
    "h6",
    "p",
    "u",
    "img",
]

ALLOWED_ATTRIBUTES = {
    "*": ["title"],
    "p": ["style"],
    "a": ["href", "title"],
    "abbr": ["title"],
    "acronym": ["title"],
    "img": ["src", "alt", "title"],
}

ALLOWED_STYLES = ["text-align"]

WIDGET_ALLOWED = {}
for attr in ALLOWED_TAGS:
    allowed = ALLOWED_ATTRIBUTES.get(attr, True)
    if not isinstance(allowed, bool):
        allowed = {tag: True for tag in allowed}
    WIDGET_ALLOWED[attr] = allowed


class EventForm(Form):
    title = StringField(label=_l("Title"), filters=(strip,), validators=[required()])

    start = DateTimeField(_l("Start"), validators=[required()])
    end = DateTimeField(_l("End"), validators=[required()])

    location = TextAreaField(label=_l("Location"), filters=(strip,))

    url = URLField(label=_l("URL"), filters=(strip,))

    description = TextAreaField(
        label=_l("Description"),
        widget=RichTextWidget(allowed_tags=WIDGET_ALLOWED),
        filters=(strip,),
        validators=[required()],
    )

    def validate_description(self, field):
        field.data = bleach.clean(
            field.data,
            tags=ALLOWED_TAGS,
            attributes=ALLOWED_ATTRIBUTES,
            styles=ALLOWED_STYLES,
            strip=True,
        )

    def validate_end(self, field):
        if self.start.data > self.end.data:
            raise ValidationError(_l("End date/time must be after start"))


EventForm.start.kwargs["raw_data"] = [" | 09:00"]
EventForm.end.kwargs["raw_data"] = [" | 18:00"]



# coding=utf-8
from __future__ import absolute_import, print_function, unicode_literals

from flask import render_template

# TEMP
ROOT = "http://localhost:5000/cmis/atompub"
XML_HEADER = "<?xml version='1.0' encoding='UTF-8'?>\n"


class Feed(object):
    def __init__(self, object, collection):
        self.object = object
        self.collection = collection

    def to_xml(self, **options):
        ctx = {
            "ROOT": ROOT,
            "object": self.object,
            "collection": self.collection,
            "to_xml": to_xml,
        }
        return render_template("cmis/feed.xml", **ctx)


class Entry(object):
    def __init__(self, obj):
        self.obj = obj

    def to_xml(self, **options):
        ctx = {
            "ROOT": ROOT,
            "folder": self.obj,
            "document": self.obj,
            "options": options,
            "to_xml": to_xml,
        }

        if self.obj.sbe_type == "cmis:folder":
            result = render_template("cmis/folder.xml", **ctx)
        elif self.obj.sbe_type == "cmis:document":
            result = render_template("cmis/document.xml", **ctx)
        else:
            raise Exception("Unknown base object type: %s" % self.obj.sbe_type)

        if "no_xml_header" not in options:
            result = XML_HEADER + result

        return result


def to_xml(obj, **options):
    entry = Entry(obj)
    return entry.to_xml(**options)



# coding=utf-8
from __future__ import absolute_import, print_function, unicode_literals

from flask import g
from markdown import Markdown
from mock import MagicMock, patch
from six import text_type
from six.moves.urllib.parse import quote_plus

from abilian.sbe.apps.wiki.markup import SBEWikiLinkExtension
from abilian.sbe.apps.wiki.models import WikiPage

pytest_plugins = ["abilian.sbe.apps.communities.tests.fixtures"]


def test_wikilink_extension():
    text = "/*€("
    wikilink = "[[" + text + "]]"

    def build_url(label, base, end):
        return "/?title=" + quote_plus(label.encode("utf-8")) + end

    extension = SBEWikiLinkExtension([("build_url", build_url)])
    ctx = {}
    ctx["extensions"] = [extension, "toc"]
    ctx["output_format"] = "html5"
    md = Markdown(**ctx)

    page_exists_mock = MagicMock(return_value=True)
    with patch("abilian.sbe.apps.wiki.forms.page_exists", page_exists_mock):
        result = md.convert(wikilink)

    qtext = text_type(quote_plus(text.encode("utf-8")))
    expected = '<p><a class="wikilink" href="/?title={href}/">{text}</a></p>'
    expected = expected.format(href=qtext, text=text)
    assert expected == result

    # make sur the class is 'wikilink new' if page exists'
    page_exists_mock = MagicMock(return_value=False)
    with patch("abilian.sbe.apps.wiki.forms.page_exists", page_exists_mock):
        result = md.convert(wikilink)

    qtext = text_type(quote_plus(text.encode("utf-8")))
    expected = '<p><a class="wikilink new" href="/?title={href}/">{text}</a></p>'
    expected = expected.format(href=qtext, text=text)
    assert expected == result


def test_new_page(user, req_ctx):
    g.user = user

    page = WikiPage(title="Some page", body_src="abc")
    assert page.title == "Some page"
    assert page.name == "Some page"
    assert page.body_src == "abc"
    assert page.body_html == "<p>abc</p>"
    assert len(page.revisions) == 1

    revision = page.revisions[0]
    assert revision.number == 0
    assert revision.author == user


def test_rename_page(user, req_ctx):
    g.user = user

    page = WikiPage(title="Some page", body_src="abc")
    assert page.title == "Some page"
    assert page.name == "Some page"

    page.title = "Title Renamed"
    assert page.title == "Title Renamed"
    assert page.name == "Title Renamed"

    page.name = "Name"
    assert page.title == "Name"
    assert page.name == "Name"


def test_create_revision(user, req_ctx):
    g.user = user

    page = WikiPage("abc")
    page.create_revision("def", "page updated")

    assert len(page.revisions) == 2
    last_revision = page.revisions[1]
    assert last_revision.number == 1
    assert last_revision.author == user



# coding=utf-8
""""""
from __future__ import absolute_import, division, print_function, \
    unicode_literals

import sys
import unittest
from io import BytesIO
from itertools import count
from pathlib import Path
from zipfile import ZipFile

import flask_mail
import pytest
from abilian.testing.util import path_from_url
from abilian.web.util import url_for
from flask import g, get_flashed_messages
from toolz import first
from werkzeug.datastructures import FileStorage

from abilian.sbe.apps.communities.models import WRITER
from abilian.sbe.apps.communities.presenters import CommunityPresenter
from abilian.sbe.apps.communities.tests.base import CommunityBaseTestCase, \
    CommunityIndexingTestCase

from ..models import Folder, PathAndSecurityIndexable, db
from ..views import util as view_util
from ..views.folders import explore_archive


def open_file(filename):
    path = Path(__file__).parent / "data" / "dummy_files" / filename
    return path.open("rb")


class BaseTests(CommunityBaseTestCase):
    no_login = True

    def setUp(self):
        super(BaseTests, self).setUp()

        self.root_folder = Folder(title="root")
        db.session.add(self.root_folder)
        db.session.flush()

    @staticmethod
    def uid_from_url(url):
        return int(url.split("/")[-1])

    @staticmethod
    def open_file(filename):
        return open_file(filename)


class TestViews(CommunityIndexingTestCase, BaseTests):
    def setUp(self):
        super(TestViews, self).setUp()
        self.community.type = "participative"
        self.community.set_membership(self.user, WRITER)
        self.session.commit()
        self.folder = self.community.folder

    def test_util_create(self):
        with self.login(self.user):
            g.user = self.user
            g.community = CommunityPresenter(self.community)
            name = "document"
            fs = FileStorage(
                BytesIO(b"content"), filename=name, content_type="text/plain"
            )
            doc = view_util.create_document(self.folder, fs)
            self.session.flush()
            assert doc.parent == self.folder
            assert doc.name == name

            # test upload with same name: should be renamed
            fs = FileStorage(
                BytesIO(b"content"), filename=name, content_type="text/plain"
            )
            doc2 = view_util.create_document(self.folder, fs)
            self.session.flush()
            assert doc2.parent == self.folder
            assert len(self.folder.children) == 2
            assert doc2.name == name + "-1"

            messages = get_flashed_messages()
            assert len(messages) == 1

    def test_home(self):
        with self.client_login(self.user.email, password="azerty"):
            response = self.get(
                url_for("documents.index", community_id=self.community.slug)
            )
            assert response.status_code == 302
            path = path_from_url(response.location)
            expected = "/communities/{}/docs/folder/{}".format(
                self.community.slug, self.folder.id
            )
            assert path == expected

    def _test_upload(
        self, title, content_type, test_preview=True, assert_preview_available=True
    ):
        data = {
            "file": (self.open_file(title), title, content_type),
            "action": "upload",
        }

        folder = self.community.folder
        url = url_for(
            "documents.folder_post",
            community_id=self.community.slug,
            folder_id=folder.id,
        )
        response = self.client.post(url, data=data)
        assert response.status_code == 302

        doc = folder.children[0]
        assert doc.title == title

        url = url_for(
            "documents.document_view", community_id=self.community.slug, doc_id=doc.id
        )
        response = self.get(url)
        assert response.status_code == 200

        url = url_for(
            "documents.document_download",
            community_id=self.community.slug,
            doc_id=doc.id,
        )
        response = self.get(url)
        assert response.status_code == 200
        assert response.headers["Content-Type"] == content_type

        content = self.open_file(title).read()
        assert response.data == content

        if test_preview:
            url = url_for(
                "documents.document_preview_image",
                community_id=self.community.slug,
                doc_id=doc.id,
                size=500,
            )
            response = self.get(url)
            if assert_preview_available:
                assert response.status_code == 200
                assert response.headers["Content-Type"] == "image/jpeg"
            else:
                # redirect to 'missing image'
                assert response.status_code == 302
                assert response.headers["Cache-Control"] == "no-cache"

        url = url_for(
            "documents.document_delete", community_id=self.community.slug, doc_id=doc.id
        )
        response = self.client.post(url)
        assert response.status_code == 302

        url = url_for(
            "documents.document_view", community_id=self.community.slug, doc_id=doc.id
        )
        response = self.get(url)
        assert response.status_code == 404

    def test_text_upload(self):
        name = "wikipedia-fr.txt"
        with self.client_login(self.user.email, password="azerty"):
            self._test_upload(name, "text/plain", test_preview=False)

    def test_pdf_upload(self):
        name = "onepage.pdf"
        with self.client_login(self.user.email, password="azerty"):
            self._test_upload(name, "application/pdf")

    def test_image_upload(self):
        NAME = "picture.jpg"
        with self.client_login(self.user.email, password="azerty"):
            self._test_upload(NAME, "image/jpeg")

    @pytest.mark.skip()  # FIXME: magic detection mismatch?
    def test_binary_upload(self):
        NAME = "random.bin"
        with self.client_login(self.user.email, password="azerty"):
            self._test_upload(
                NAME, "application/octet-stream", assert_preview_available=False
            )

    @pytest.mark.skipif(sys.version_info >= (3, 0), reason="Doesn't work yet on Py3k")
    def test_explore_archive(self):
        fd = self.open_file("content.zip")
        result = [("/".join(path), f) for path, f in explore_archive(fd)]
        assert result == [("", fd)]

        fd = self.open_file("content.zip")
        archive_content = explore_archive(fd, uncompress=True)
        result = {"/".join(path) + "/" + f.filename for path, f in archive_content}
        assert result == {
            "existing-doc/file.txt",
            "existing-doc/subfolder_in_renamed/doc.txt",
            "folder 1/doc.txt",
            "folder 1/dos cp437: é.txt",
            "folder 1/osx: utf-8: é.txt",
        }

    def test_zip_upload_uncompress(self):
        folder = Folder(title="folder 1", parent=self.community.folder)
        self.session.add(folder)
        self.session.flush()
        folder = self.community.folder
        files = []
        files.append((BytesIO(b"A document"), "existing-doc", "text/plain"))
        files.append((self.open_file("content.zip"), "content.zip", "application/zip"))
        data = {"file": files, "action": "upload", "uncompress_files": True}
        url = url_for(
            "documents.folder_post",
            community_id=self.community.slug,
            folder_id=folder.id,
        )
        with self.client_login(self.user.email, password="azerty"):
            response = self.client.post(url, data=data)

        assert response.status_code == 302
        expected = {"existing-doc", "folder 1", "existing-doc-1"}
        assert expected == {f.title for f in folder.children}
        expected = {"folder 1", "existing-doc-1"}
        assert expected == {f.title for f in folder.subfolders}

    def test_zip(self):
        with self.client_login(self.user.email, password="azerty"):
            title = "onepage.pdf"
            content_type = "application/pdf"
            data = {
                "file": (self.open_file(title), title, content_type),
                "action": "upload",
            }
            folder = self.community.folder
            url = url_for(
                "documents.folder_post",
                community_id=self.community.slug,
                folder_id=folder.id,
            )
            response = self.client.post(url, data=data)
            assert response.status_code == 302

            doc = folder.children[0]
            data = {
                "action": "download",
                "object-selected": ["cmis:document:%d" % doc.id],
            }
            url = url_for(
                "documents.folder_post",
                community_id=self.community.slug,
                folder_id=folder.id,
            )
            response = self.client.post(url, data=data)
            assert response.status_code == 200
            assert response.content_type == "application/zip"

            zipfile = ZipFile(BytesIO(response.data))
            assert [zipfile.namelist()[0]] == [title]

    def test_recursive_zip(self):
        with self.client_login(self.user.email, password="azerty"):
            data = {"action": "new", "title": "my folder"}
            folder = self.community.folder
            url = url_for(
                "documents.folder_post",
                community_id=self.community.slug,
                folder_id=folder.id,
            )
            response = self.client.post(url, data=data)
            assert response.status_code == 302

            my_folder = folder.children[0]

            title = "onepage.pdf"
            content_type = "application/pdf"
            data = {
                "file": (self.open_file(title), title, content_type),
                "action": "upload",
            }
            url = url_for(
                "documents.folder_post",
                community_id=self.community.slug,
                folder_id=my_folder.id,
            )
            response = self.client.post(url, data=data)
            assert response.status_code == 302

            data = {
                "action": "download",
                "object-selected": ["cmis:folder:%d" % my_folder.id],
            }
            url = url_for(
                "documents.folder_post",
                community_id=self.community.slug,
                folder_id=folder.id,
            )
            response = self.client.post(url, data=data)
            assert response.status_code == 200
            assert response.content_type == "application/zip"

            zipfile = ZipFile(BytesIO(response.data))
            assert zipfile.namelist() == ["my folder/" + title]

    def test_document_send_by_mail(self):
        mail = self.app.extensions["mail"]
        folder = self.community.folder

        with self.client_login(self.user.email, password="azerty"):
            # upload files
            for filename in ("ascii title.txt", "utf-8 est arrivé!.txt"):
                content_type = "text/plain"
                data = {
                    "file": (BytesIO(b"file content"), filename, content_type),
                    "action": "upload",
                }
                url = url_for(
                    "documents.folder_post",
                    community_id=self.community.slug,
                    folder_id=folder.id,
                )
                self.client.post(url, data=data)

            ascii_doc = folder.children[0]
            unicode_doc = folder.children[1]

            def get_send_url(doc_id):
                return url_for(
                    "documents.document_send",
                    community_id=self.community.slug,
                    doc_id=doc_id,
                )

            # mail ascii filename
            with mail.record_messages() as outbox:
                url = get_send_url(ascii_doc.id)
                response = self.client.post(
                    url,
                    data={
                        "recipient": "dest@example.com",
                        "message": "Voilà un fichier",
                    },
                )
                assert response.status_code == 302
                assert len(outbox) == 1

                msg = outbox[0]
                assert msg.subject == "[Abilian Test] Unknown sent you a file"
                assert msg.recipients == ["dest@example.com"]

                assert len(msg.attachments) == 1

                attachment = first(msg.attachments)
                assert isinstance(attachment, flask_mail.Attachment)
                assert attachment.filename == "ascii title.txt"

            # mail unicode filename
            with mail.record_messages() as outbox:
                url = get_send_url(unicode_doc.id)
                response = self.client.post(
                    url,
                    data={
                        "recipient": "dest@example.com",
                        "message": "Voilà un fichier",
                    },
                )
                assert response.status_code == 302
                assert len(outbox) == 1

                msg = outbox[0]
                assert isinstance(msg, flask_mail.Message)
                assert msg.subject == "[Abilian Test] Unknown sent you a file"
                assert msg.recipients == ["dest@example.com"]
                assert len(msg.attachments) == 1

                attachment = first(msg.attachments)
                assert isinstance(attachment, flask_mail.Attachment)
                assert attachment.filename == "utf-8 est arrivé!.txt"


class TestPathIndexable(unittest.TestCase):
    class MockPath(PathAndSecurityIndexable):
        def __init__(self, id, parent=None):
            self.id = id
            self.parent = parent

    def setUp(self):
        id_gen = count()
        obj = self.MockPath(next(id_gen))
        obj = self.MockPath(next(id_gen), parent=obj)
        obj = self.MockPath(next(id_gen), parent=obj)
        self.obj = self.MockPath(next(id_gen), parent=obj)

    def test_iter_to_root(self):
        assert [o.id for o in self.obj._iter_to_root()] == [3, 2, 1, 0]
        assert [o.id for o in self.obj._iter_to_root(skip_self=True)] == [2, 1, 0]

    def test_indexable_parent_ids(self):
        assert self.obj._indexable_parent_ids == "/0/1/2"



# coding=utf-8
from __future__ import absolute_import, print_function, unicode_literals

import re

from abilian.core.models.subjects import User
from flask import g, url_for
from toolz import first

from abilian.sbe.apps.communities.tests.base import CommunityBaseTestCase, \
    CommunityIndexingTestCase
from abilian.sbe.apps.wiki.models import WikiPage


class TestIndexing(CommunityIndexingTestCase):
    def setUp(self):
        super(TestIndexing, self).setUp()
        g.user = User.query.all()[0]

    def test_wiki_indexed(self):
        page = WikiPage(title="Community 1", community=self.community)
        self.session.add(page)

        page_other = WikiPage(title="Community 2: other", community=self.c2)
        self.session.add(page_other)

        self.session.commit()

        svc = self.svc
        obj_types = (WikiPage.entity_type,)
        with self.login(self.user_no_community):
            res = svc.search("community", object_types=obj_types)
            assert len(res) == 0

        with self.login(self.user):
            res = svc.search("community", object_types=obj_types)
            assert len(res) == 1

            hit = res[0]
            assert hit["object_key"] == page.object_key

        with self.login(self.user_c2):
            res = svc.search("community", object_types=obj_types)
            assert len(res) == 1

            hit = res[0]
            assert hit["object_key"] == page_other.object_key


class TestsViews(CommunityBaseTestCase):
    def setUp(self):
        super(TestsViews, self).setUp()
        g.user = User.query.all()[0]

    def test_create_page(self):
        title = "Some page name"
        url = url_for("wiki.page_new", community_id=self.community.slug)
        url += "?title=Some+page+name"
        response = self.client.get(url)
        assert response.status_code == 200
        # make sure the title is filled when comming from wikilink
        assert 'value="Some page name"' in response.get_data(as_text=True)

        title = "Some page"
        body = "LuuP3jai"
        url = url_for("wiki.page_new", community_id=self.community.slug)
        data = {"title": title, "body_src": body, "__action": "create"}
        response = self.client.post(url, data=data)
        assert response.status_code == 302

        url = url_for("wiki.page", community_id=self.community.slug, title=title)
        response = self.client.get(url)
        assert response.status_code == 200
        assert body in response.get_data(as_text=True)

        # edit
        url = url_for("wiki.page_edit", community_id=self.community.slug, title=title)
        response = self.client.get(url)
        assert response.status_code == 200

        # Slightly hackish way to get the page_id
        line = first(
            l
            for l in response.get_data(as_text=True).split("\n")
            if 'name="page_id"' in l
        )
        m = re.search('value="(.*?)"', line)
        page_id = int(m.group(1))

        url = url_for(
            "wiki.page_changes", community_id=self.community.slug, title=title
        )
        response = self.client.get(url)
        assert response.status_code == 200

        url = url_for("wiki.page_source", community_id=self.community.slug, title=title)
        response = self.client.get(url)
        assert response.status_code == 200

        url = url_for("wiki.page_edit", community_id=self.community.slug)
        data = {"title": title, "page_id": page_id, "body_src": "abc def"}
        data["__action"] = "edit"
        response = self.client.post(url, data=data)
        assert response.status_code == 302

        url = url_for(
            "wiki.page_compare",
            rev0="on",
            rev1="on",
            community_id=self.community.slug,
            title=title,
        )
        response = self.client.get(url)
        assert response.status_code == 200



# coding=utf-8
""""""
from __future__ import absolute_import, print_function, unicode_literals

import sqlalchemy as sa
from flask_script import Manager

from . import tasks
from .models import Document

manager = Manager(description="SBE documents actions", help="SBE documents actions")


@manager.command
def antivirus():
    """Schedule documents to antivirus scan."""

    documents = Document.query.filter(Document.content_blob != None).options(
        sa.orm.noload("creator"),
        sa.orm.noload("owner"),
        sa.orm.joinedload("content_blob"),
    )

    total = 0
    count = 0
    for d in documents.yield_per(1000):
        total += 1
        meta = d.content_blob.meta
        if "antivirus" not in meta and "antivirus_task" not in meta:
            tasks.antivirus_scan.delay(d.id)
            count += 1

    print("{count}/{total} documents scheduled".format(count=count, total=total))



# coding=utf-8
"""Models for the Forum.

Note: a few features are planned but not implemented yet, and are commented
out.
"""
from __future__ import absolute_import, print_function, unicode_literals

from collections import Counter
from datetime import datetime
from itertools import chain

from abilian.core.entities import SEARCHABLE, Entity
from abilian.services.indexing.adapter import SAAdapter
from sqlalchemy import Column, ForeignKey, Integer, Unicode, UnicodeText
from sqlalchemy.event import listens_for
from sqlalchemy.ext.hybrid import hybrid_property
from sqlalchemy.orm import backref, relationship
from sqlalchemy.types import DateTime

from abilian.sbe.apps.communities.models import Community, CommunityIdColumn, \
    community_content
from abilian.sbe.apps.documents.models import BaseContent, CmisObject


class ThreadClosedError(RuntimeError):
    def __init__(self, thread):
        super(ThreadClosedError, self).__init__(
            "The thread {!r} is closed. No modification allowed on its posts: "
            "creation, edition, deletion".format(thread)
        )
        self.thread = thread


@community_content
class Thread(Entity):
    """A thread contains conversations among forum participants.

    The discussions in a thread may be sorted in chronological order or threaded
    by reply.

    (= Thread in SIOC, Message in ICOM 1.0).
    """

    __tablename__ = "forum_thread"

    community_id = CommunityIdColumn()
    #: The community this thread belongs to
    community = relationship(
        Community,
        primaryjoin=(community_id == Community.id),
        backref=backref("threads", cascade="all, delete-orphan"),
    )

    #: The thread title (aka subject)
    _title = Column("title", Unicode(255), nullable=False, default="", info=SEARCHABLE)

    last_post_at = Column(DateTime, default=datetime.utcnow, nullable=True)

    # title is defined has an hybrid property to allow name <-> title sync (2
    # way)
    @hybrid_property
    def title(self):
        return self._title

    def get_frequent_posters(self, limit):
        all_posts = self.posts[1:]
        posters_counter = Counter([e.creator for e in all_posts])
        sorted_posters = posters_counter.most_common(limit)
        frequent_posters = [
            user for (user, nb_posts) in sorted_posters if user != self.creator
        ]
        return frequent_posters

    @title.setter
    def title(self, title):
        # set title before setting name, so that we don't enter an infinite loop
        # with _thread_sync_name_title
        self._title = title
        if self.name != title:
            self.name = title

    posts = relationship(
        "Post",
        primaryjoin="Thread.id == Post.thread_id",
        order_by="Post.created_at",
        cascade="all, delete-orphan",
        back_populates="thread",
    )

    @property
    def closed(self):
        """True if this thread doesn't accept more posts."""
        return self.meta.get("abilian.sbe.forum", {}).get("closed", False)

    @closed.setter
    def closed(self, value):
        self.meta.setdefault("abilian.sbe.forum", {})["closed"] = bool(value)
        self.meta.changed()

    def create_post(self, **kw):
        if self.closed:
            raise ThreadClosedError(self)

        kw["name"] = self.name
        post = Post(**kw)
        post.thread = self
        return post


@listens_for(Thread.name, "set", active_history=True)
def _thread_sync_name_title(entity, new_value, old_value, initiator):
    """Synchronize thread name -> title.

    thread.title -> name is done via hybrid_property, avoiding infinite
    loop (since "set" is received before attribute has received value)
    """
    if entity.title != new_value:
        entity.title = new_value
    return new_value


class Post(Entity):
    """A post is a message in a forum discussion thread.

    (= Post in DiscussionMessage in ICOM 1.0).
    """

    __tablename__ = "forum_post"
    __indexable__ = False  # content is indexed at thread level

    #: The thread this post belongs to
    thread_id = Column(ForeignKey(Thread.id), nullable=False)
    thread = relationship(Thread, foreign_keys=thread_id, back_populates="posts")

    #: The post this post is a reply to, if any (currently not used)
    parent_post_id = Column(ForeignKey("forum_post.id"), nullable=True)
    parent_post = relationship("Post", foreign_keys=[parent_post_id])

    #: Markup type (Markdown, Textile...)
    # TODO: markup langage selection + default
    # markup_type = Column(String, default="Markdown")

    #: Source (markup) for the post
    # body_src = Column(UnicodeText, default=u"", nullable=False)

    #: HTML rendering of the post
    body_html = Column(UnicodeText, default="", nullable=False)

    @hybrid_property
    def title(self):
        return self.name

    @property
    def history(self):
        return self.meta.get("abilian.sbe.forum", {}).get("history", [])


class ThreadIndexAdapter(SAAdapter):
    """Index a thread and its posts."""

    @staticmethod
    def can_adapt(obj_cls):
        return obj_cls is Thread

    def get_document(self, obj):
        kw = super(ThreadIndexAdapter, self).get_document(obj)
        kw["text"] = " ".join(chain((kw["text"],), [p.body_html for p in obj.posts]))
        return kw


# event listener to sync name with thread's name
@listens_for(Thread.name, "set", active_history=True)
def _thread_sync_name(thread, new_value, old_value, initiator):
    """Synchronize name with thread's name."""
    if new_value == old_value:
        return new_value

    for post in thread.posts:
        post.name = new_value
    return new_value


@listens_for(Post.thread, "set", active_history=True)
def _thread_change_sync_name(post, new_thread, old_thread, initiator):
    """Change name on thread change."""
    if new_thread == old_thread or new_thread is None:
        return new_thread
    post.name = new_thread.name
    return new_thread


@listens_for(Thread.posts, "append")
@listens_for(Thread.posts, "remove")
@listens_for(Thread.posts, "set")
def _guard_closed_thread_collection(thread, value, *args):
    """Prevent add/remove/replace posts on a closed thread."""
    if isinstance(thread, Post):
        thread = thread.thread
        if thread is None:
            return

    if thread.closed:
        raise ThreadClosedError(thread)

    return value


class PostAttachment(BaseContent, CmisObject):
    __tablename__ = None  # type: str
    __mapper_args__ = {"polymorphic_identity": "forum_post_attachment"}
    sbe_type = "forum_post:attachment"

    _post_id = Column(Integer, ForeignKey(Post.id), nullable=True)
    post = relationship(
        Post,
        primaryjoin=(_post_id == Post.id),
        backref=backref(
            "attachments",
            lazy="select",
            order_by="PostAttachment.name",
            cascade="all, delete-orphan",
        ),
    )



# coding=utf-8
from __future__ import absolute_import, division, print_function, \
    unicode_literals

from pathlib import Path

from lxml import etree

from ..webdav.constants import DAV_PROPS
from ..webdav.xml import MultiStatus, Propfind


def test_propfind_sample1():
    xml = (Path(__file__).parent / "data" / "propfind1.xml").open("rb").read()
    propfind = Propfind(xml)
    assert propfind.mode == "prop"


def test_propfind_sample2():
    xml = (Path(__file__).parent / "data" / "propfind2.xml").open("rb").read()
    propfind = Propfind(xml)
    assert propfind.mode == "prop"


def test_propfind_sample3():
    xml = (Path(__file__).parent / "data" / "propfind3.xml").open("rb").read()
    propfind = Propfind(xml)
    assert propfind.mode == "allprop"


def test_empty_multistatus():
    m = MultiStatus()
    result = m.to_string()

    # Check XML is weel-formed
    etree.fromstring(result)


def test_multistatus():
    class Obj(object):
        pass

    obj = Obj()
    obj.name = "some name"
    obj.is_folder = True
    m = MultiStatus()
    m.add_response_for("http://example.com/", obj, DAV_PROPS)
    result = m.to_string()

    # Check XML is weel-formed
    etree.fromstring(result)



# coding=utf-8

from __future__ import absolute_import, print_function, unicode_literals

from abilian.i18n import _l
from abilian.services import get_service
from abilian.services.security import Admin
from abilian.web.action import Action, FAIcon, actions
from flask import g, url_for
from flask_login import current_user


class CalendarAction(Action):
    def url(self, context=None):
        return url_for("." + self.name, community_id=g.community.slug)


class EventAction(CalendarAction):
    def pre_condition(self, context):
        event = context.get("object")
        return not not event

    def url(self, context=None):
        event = context.get("object")
        return url_for(
            "." + self.name, community_id=g.community.slug, event_id=event.id
        )


def is_admin(context):
    security = get_service("security")
    return security.has_role(current_user, Admin, object=context.get("object"))


_actions = [
    CalendarAction(
        "calendar:global", "new_event", _l("Create a new event"), icon="plus"
    ),
    CalendarAction("calendar:global", "index", _l("Upcoming events"), icon="list"),
    EventAction("calendar:event", "event", _l("View event"), icon=FAIcon("eye")),
    EventAction(
        "calendar:event", "event_edit", _l("Edit event"), icon=FAIcon("pencil")
    ),
]


def register_actions(state):
    if not actions.installed(state.app):
        return
    with state.app.app_context():
        actions.register(*_actions)



# coding=utf-8
"""Parses and produces XML documents specified by the standard."""

from __future__ import absolute_import, print_function, unicode_literals

from lxml import etree, objectify
from lxml.builder import ElementMaker

E = ElementMaker(namespace="DAV:")


class Propfind(object):
    def __init__(self, xml=""):
        self.mode = ""
        self.prop_names = []

        if not xml:
            xml = "<D:propfind xmlns:D='DAV:'><D:allprop/></D:propfind>"
        self.parse(xml)

    def parse(self, xml):
        root = objectify.fromstring(xml)

        child = root.getchildren()[0]
        self.mode = child.tag[len("{DAV:}") :]

        if self.mode == "prop":
            for prop in child.getchildren():
                self.prop_names.append(prop.tag)


class MultiStatus(object):
    def __init__(self):
        self.responses = []

    def add_response_for(self, href, obj, property_list):
        response = Response(href, obj, property_list)
        self.responses.append(response)

    def to_string(self):
        return etree.tostring(self.to_xml(), pretty_print=True)

    def to_xml(self):
        xml = E.multistatus()
        for response in self.responses:
            xml.append(response.to_xml())
        return xml


class Response(object):
    def __init__(self, href, obj, property_list):
        self.href = href
        self.property_list = property_list
        self.obj = obj

    def to_xml(self):
        obj = self.obj
        props = E.prop()
        for property_name in self.property_list:
            if property_name == "creationdate":
                props.append(E.creationdate("1997-12-01T17:42:21-08:00"))
            elif property_name == "displayname":
                props.append(E.displayname(obj.name))
            elif property_name == "resourcetype":
                if obj.is_folder:
                    props.append(E.resourcetype(E.collection()))
                else:
                    props.append(E.resourcetype())

        return E.response(
            E.href(self.href), E.propstat(props), E.status("HTTP/1.1 200 OK")
        )



# coding=utf-8
from __future__ import absolute_import

from flask import Blueprint

notifications = Blueprint(
    "notifications",
    __name__,
    url_prefix="/notifications",
    template_folder="../templates",
)



SITE_NAME = "Abilian SBE Dev"
TESTING = False
PRODUCTION = False
CSRF_ENABLED = True
SECRET_KEY = 'CHANGEME'

PREFERRED_URL_SCHEME = 'http'

DEBUG = True
DEBUG_TOOLBAR = True
DEBUG_TB_INTERCEPT_REDIRECTS = False
DEBUG_TB_PROFILER_ENABLED = True

LOGGING_CONFIG_FILE = 'logging.yml'

SQLALCHEMY_DATABASE_URI = "sqlite:///INSTANCE_FOLDER/data.db"
# SQLALCHEMY_DATABASE_URI = postgres://user@localhost:5432/abilian_sbe
# SQLALCHEMY_ECHO = True

# run tasks locally, no async
CELERY_ALWAYS_EAGER = True
CELERY_EAGER_PROPAGATES_EXCEPTIONS = True

MAIL_SENDER = "user@localhost"
MAIL_FAIL_SILENTLY = False



#!./bin/python
# coding=utf-8

from __future__ import absolute_import, print_function, unicode_literals

import subprocess
import time

from six.moves.urllib.parse import urlopen

# Some random number
BIND = "0.0.0.0"
PORT = 4034
HOME = "http://{}:{}/".format(BIND, PORT)

subprocess.call(["./bin/pip", "install", "gunicorn"])

p = subprocess.Popen(["./bin/gunicorn", "wsgi:app", "-b", "{}:{}".format(BIND, PORT)])

try:
    # Just in case
    time.sleep(5)
    page = urlopen(HOME).read()
    assert "Welcome to Abilian" in page
finally:
    p.terminate()
    p.wait()



# coding=utf-8
"""Forum module."""
from __future__ import absolute_import

from abilian.sbe.extension import sbe


def register_plugin(app):
    app.config.setdefault("SBE_FORUM_REPLY_BY_MAIL", False)
    app.config.setdefault("MAIL_ADDRESS_TAG_CHAR", u"+")
    app.config.setdefault("INCOMING_MAIL_USE_MAILDIR", False)
    sbe.init_app(app)
    from .views import forum
    from .actions import register_actions
    from .models import ThreadIndexAdapter
    from .commands import manager
    from . import tasks

    forum.record_once(register_actions)
    app.register_blueprint(forum)
    app.services["indexing"].adapters_cls.insert(0, ThreadIndexAdapter)
    tasks.init_app(app)

    if app.script_manager:
        app.script_manager.add_command("forum", manager)



# coding=utf-8
"""The blueprint for this app."""

from __future__ import absolute_import, print_function, unicode_literals

from abilian.core.extensions import db
from abilian.core.util import get_params
from flask import Blueprint, redirect, render_template, url_for

from abilian.sbe.apps.wall.presenters import ActivityEntryPresenter
from abilian.sbe.apps.wall.util import get_recent_entries

from ..models import Message

__all__ = ["social"]

MAX_LAST_USERS = 15
MAX_MESSAGES = 15

social = Blueprint(
    "social", __name__, url_prefix="/social", template_folder="../templates"
)
route = social.route


@route("/")
def home():
    ctx = {}
    entries = get_recent_entries(num=50)
    entries = ActivityEntryPresenter.wrap_collection(entries)
    ctx["activity_entries"] = entries

    return render_template("social/home.html", **ctx)


@route("/stream/<stream_name>")
def stream(stream_name):
    pass


@route("/", methods=["POST"])
def share():
    # TODO: better error control / feedback
    d = get_params(Message.__editable__)
    if not d.get("content"):
        return redirect(url_for(".home"))

    message = Message(**d)
    db.session.add(message)

    # tags = message.tags
    # values = [ {'tag': tag, 'message_id': message.id} for tag in tags ]
    # db.engine.execute(tagging.insert(), values)

    db.session.commit()
    return redirect(url_for(".home"))



# coding=utf-8
""""""
from __future__ import absolute_import, print_function, unicode_literals

import bleach
from abilian.i18n import _l
from abilian.web.forms import Form, RichTextWidget
from abilian.web.forms.fields import FileField
from abilian.web.forms.filters import strip
from abilian.web.forms.validators import Length, optional, required
from wtforms import BooleanField, StringField, TextAreaField

ALLOWED_TAGS = [
    "a",
    "abbr",
    "acronym",
    "b",
    "blockquote",
    "br",
    "code",
    "em",
    "i",
    "li",
    "ol",
    "strong",
    "ul",
    "h1",
    "h2",
    "h3",
    "h4",
    "h5",
    "h6",
    "p",
    "u",
    "img",
]

ALLOWED_ATTRIBUTES = {
    "*": ["title"],
    "p": ["style"],
    "a": ["href", "title"],
    "abbr": ["title"],
    "acronym": ["title"],
    "img": ["src", "alt", "title"],
}

ALLOWED_STYLES = ["text-align"]

WIDGET_ALLOWED = {}
for attr in ALLOWED_TAGS:
    allowed = ALLOWED_ATTRIBUTES.get(attr, True)
    if not isinstance(allowed, bool):
        allowed = {tag: True for tag in allowed}
    WIDGET_ALLOWED[attr] = allowed

# instantiate this one before PostForm fields, so that it is listed first
# when Threadform is displayed
_TITLE_FIELD = StringField(
    label=_l("Title"), filters=(strip,), validators=[required(), Length(max=150)]
)


class BasePostForm(Form):
    message = TextAreaField(
        label=_l("Message"),
        widget=RichTextWidget(allowed_tags=WIDGET_ALLOWED),
        filters=(strip,),
        validators=[required()],
    )
    attachments = FileField(
        label=_l("Attachments"), multiple=True, validators=[optional()]
    )

    def validate_message(self, field):
        field.data = bleach.clean(
            field.data,
            tags=ALLOWED_TAGS,
            attributes=ALLOWED_ATTRIBUTES,
            styles=ALLOWED_STYLES,
            strip=True,
        )


class PostForm(BasePostForm):
    send_by_email = BooleanField(label=_l("Send by email?"), default=True)


class ThreadForm(PostForm):
    title = _TITLE_FIELD


class PostEditForm(BasePostForm):
    reason = StringField(
        label=_l("Reason"),
        description=_l("Description of your edit"),
        filters=(strip,),
        validators=(optional(),),
    )



# coding=utf-8
"""Register extensions as a plugin.

NOTE: panels are currently loaded and registered manually. This may change
in the future.
"""
from __future__ import absolute_import

from abilian.sbe.extension import sbe


def register_plugin(app):
    sbe.init_app(app)
    from .main import blueprint

    app.register_blueprint(blueprint)



# coding=utf-8
from __future__ import absolute_import, print_function, unicode_literals

from abilian.services import get_service
from abilian.services.security import Admin
from abilian.web.action import Action, FAIcon, ModalActionMixin, actions
from flask import url_for
from flask_babel import lazy_gettext as _l
from flask_login import current_user

from abilian.sbe.apps.communities.actions import CommunityEndpoint
from abilian.sbe.apps.communities.security import is_manager


class WikiPageAction(Action):
    Endpoint = CommunityEndpoint

    def pre_condition(self, context):
        page = context.get("object")
        return bool(page)

    def url(self, context=None):
        if self._url:
            return self._url
        else:
            page = context.get("object")
            kw = self.endpoint.get_kwargs()
            kw["title"] = page.title
            return url_for(self.endpoint.name, **kw)


def is_admin(context):
    security = get_service("security")
    return security.has_role(current_user, Admin, object=context.get("object"))


class WikiPageModalAction(ModalActionMixin, WikiPageAction):
    pass


class WikiAction(Action):
    Endpoint = CommunityEndpoint


_actions = (
    WikiPageAction(
        "wiki:page",
        "page_viewers",
        _l("Readers list"),
        icon="user",
        condition=lambda ctx: is_manager(context=ctx),
        endpoint=".page_viewers",
    ),
    WikiPageAction("wiki:page", "view", _l("View"), endpoint=".page", icon="eye-open"),
    WikiPageAction(
        "wiki:page", "edit", _l("Edit"), endpoint=".page_edit", icon="pencil"
    ),
    WikiPageModalAction(
        "wiki:page",
        "upload_attachment",
        _l("Upload an attachment"),
        url="#upload-files",
        icon="plus",
    ),
    WikiPageAction(
        "wiki:page",
        "source",
        _l("Source"),
        endpoint=".page_source",
        icon=FAIcon("code"),
    ),
    WikiPageAction(
        "wiki:page", "changes", _l("Changes"), endpoint=".page_changes", icon="time"
    ),
    WikiPageModalAction(
        "wiki:page", "delete", _l("Delete"), url="#modal-delete", icon="trash"
    ),
    WikiAction("wiki:global", "new", _l("New page"), endpoint=".page_new", icon="plus"),
    WikiAction(
        "wiki:global", "pages", _l("All pages"), endpoint=".wiki_pages", icon="list"
    ),
    WikiAction(
        "wiki:global",
        "help",
        _l("Syntax help"),
        endpoint=".wiki_help",
        icon="info-sign",
    ),
)


def register_actions(state):
    if not actions.installed(state.app):
        return

    with state.app.app_context():
        actions.register(*_actions)



# coding=utf-8
from __future__ import absolute_import, print_function, unicode_literals

from flask import url_for


def test_sbe_notifications(client, login_user, req_ctx):
    response = client.get(url_for("preferences.sbe_notifications"))
    assert response.status_code == 200



# coding=utf-8
"""Configuration and injectable fixtures for Pytest.

Reuses fixtures defined in abilian-core.
"""
from __future__ import absolute_import, division, print_function, \
    unicode_literals

from pytest import fixture

from abilian.sbe.app import create_app

pytest_plugins = ["abilian.testing.fixtures"]


@fixture
def app(config):
    """Return an App configured with config=TestConfig."""
    return create_app(config=config)



# coding=utf-8
""""""
from __future__ import absolute_import, print_function, unicode_literals

from abilian.core.models.subjects import User
from abilian.web import url_for
from flask import render_template

from abilian.sbe.apps.communities.models import WRITER, Community
from abilian.sbe.apps.notifications.tasks.social import CommunityDigest, \
    generate_unsubscribe_token


def test_unsubscribe(app, client, db, app_context):
    user = User(email="user_1@example.com", password="abc", can_login=True)
    db.session.add(user)
    db.session.commit()

    preferences = app.services["preferences"]
    preferences.set_preferences(user, **{"sbe:notifications:daily": True})
    token = generate_unsubscribe_token(user)
    url = url_for("notifications.unsubscribe_sbe", token=token)

    # Not need to login, since we're using the unsubscribe token

    response = client.get(url)
    assert response.status_code == 200
    prefs = preferences.get_preferences(user)
    assert prefs["sbe:notifications:daily"]

    response = client.post(url)
    assert response.status_code == 200
    prefs = preferences.get_preferences(user)
    assert not prefs["sbe:notifications:daily"]


def test_mail_templates(db, app_context):
    # this actually tests that templates are parsed without errors, not the
    # rendered content
    user = User(email="user_1@example.com")
    db.session.add(user)
    community = Community(name="My Community")
    db.session.add(community)
    community.set_membership(user, WRITER)
    db.session.commit()

    digests = [CommunityDigest(community)]
    token = generate_unsubscribe_token(user)
    ctx = {"digests": digests, "token": token}
    render_template("notifications/daily-social-digest.txt", **ctx)
    render_template("notifications/daily-social-digest.html", **ctx)



# coding=utf-8
""""""
from __future__ import absolute_import, print_function, unicode_literals

import logging

import bleach
from abilian.core.util import BasePresenter
from abilian.i18n import _, _l
from abilian.web.util import url_for
from flask import render_template_string
from jinja2 import Markup, Template
from singledispatch import singledispatch

from abilian.sbe.apps.calendar.models import Event
from abilian.sbe.apps.communities.models import Community
from abilian.sbe.apps.documents.models import Document
from abilian.sbe.apps.forum.models import Post, Thread
from abilian.sbe.apps.wiki.models import WikiPage

logger = logging.getLogger(__name__)

# Poor man's pattern matching.
MESSAGES = {
    ("post", Thread): _l('has started conversation "{object}"'),
    ("post", Post): _l('has participated in conversation "{object}"'),
    ("post", Document): _l('has published document "{object}"'),
    ("post", WikiPage): _l('has created wiki page "{object}"'),
    ("post", Event): _l('has published event "{object}"'),
    ("post", Community): _l('has created communauté "{object}"'),
    #
    ("update", Document): _l('has updated document "{object}"'),
    ("update", WikiPage): _l('has updated wiki page "{object}"'),
    ("update", Community): _l('has updated community "{object}"'),
    #
    ("join", Community): _l("has joined the community {object}."),
    #
    ("leave", Community): _l("has left the community {object}."),
}

OBJ_TEMPLATE = Template('<a href="{{ object_url|safe }}">{{ object_name }}</a>')

POST_BODY_TEMPLATE = """
  <span class="body">"<a href="{{ object_url |safe }}">{{ body }}</a>"
  {%- if post.attachments %}
  <div id="attachments">
    <ul>
      {%- for attachment in post.attachments %}
      <li>
        <span class="attachment-item">
          <img src="{{ attachment.icon }}""
              alt=""/>
          <a href="{{ url_for(attachment) }}">{{ attachment.name }}</a>
          ({{ attachment.content_length|filesize }})
        </span>
      </li>
      {%- endfor %}
    </ul>
  </div>
  {%- endif %}
  </span>
"""

DOCUMENT_BODY_TEMPLATE = """
<div class="body">
  <div>
    <img src="{{ obj.icon }}" style="vertical-align: middle;" alt=""/>
    {% for p in parents[:-1] %}
    <a href="{{ p.path }}">{{ p.label }}</a> <span class="divider">/</span>
    {% endfor %}

    <a href="{{ url_for(obj) }}">{{ obj.name }}</a>
  </div>
  <div>
  {%- if obj.antivirus_ok %}
    <a href="{{ url_for('documents.document_download',
                         community_id=obj.community.slug,
                         doc_id=obj.id,
                         attach=True) }}">
      <i class="glyphicon glyphicon-download"></i>
      {{ _('Download') }} (<small>{{ obj.content_length|filesize  }}</small>)
    </a>
  {%- endif %}
  </div>
</div>
"""


class ActivityEntryPresenter(BasePresenter):
    @property
    def object_url(self):
        return url_for(self._model.object)

    def message(self, ignore_community=False):
        try:
            # another quick&dirty approach for now. FIXME later.
            entry = self._model
            object_class = entry.object_type.split(".")[-1]
            object_class_localized = _(object_class)

            ctx = {}
            ctx["verb"] = entry.verb

            ctx["object_name"] = entry.object.name or getattr(
                entry.object, "title", "???"
            )
            ctx["object_url"] = url_for(entry.object)
            ctx["object_type"] = object_class_localized
            ctx["object"] = OBJ_TEMPLATE.render(**ctx)

            if entry.target:
                ctx["target_name"] = entry.target.name
                ctx["target_url"] = url_for(entry.target)
                ctx["target"] = OBJ_TEMPLATE.render(
                    object_name=ctx["target_name"], object_url=ctx["target_url"]
                )

            msg = MESSAGES.get((entry.verb, entry.object.__class__))
            if msg:
                msg = msg.format(**ctx)
                if entry.target and not ignore_community:
                    msg += " " + _("in the community {target}.").format(**ctx)
                else:
                    msg += "."

            elif entry.verb == "post":
                msg = _(
                    "has posted an object of type {object_type} " 'called "{object}"'
                ).format(**ctx)

                if entry.target and not ignore_community:
                    msg += " " + _("in the community {target}.").format(**ctx)
                else:
                    msg += "."

            elif entry.verb == "join":
                msg = _("has joined the community {object}.").format(**ctx)

            elif entry.verb == "leave":
                msg = _("has left the community {object}.").format(**ctx)

            elif entry.verb == "update":
                msg = _("has updated {object_type} {object}.").format(**ctx)

            else:
                msg = _('has done action "{verb}" on object "{object}".').format(**ctx)

            return Markup(msg)

        except BaseException:
            logger.exception("Exception while presenting activity message")
            raise

    @property
    def body(self):
        return get_body(self.object)


@singledispatch
def get_body(object):
    return ""


@get_body.register(Thread)
def get_body_thread(object):
    # type: (Thread) -> Markup
    body = bleach.clean(object.posts[0].body_html, tags=[], strip=True)
    body = Markup(body).unescape()
    if len(body) > 400:
        body = body[0:400] + "…"
    body = render_template_string(
        POST_BODY_TEMPLATE, object_url=url_for(object), body=body, post=object.posts[0]
    )
    return Markup(body)


@get_body.register(Post)
def get_body_post(object):
    # type: (Post) -> Markup
    body = bleach.clean(object.body_html, tags=[], strip=True)
    body = Markup(body).unescape()
    if len(body) > 400:
        body = body[0:400] + "…"
    body = render_template_string(
        POST_BODY_TEMPLATE, object_url=url_for(object), body=body, post=object
    )
    return Markup(body)


@get_body.register(Document)
def get_body_document(object):
    # type: (Document) -> Markup
    body = bleach.clean(object.body_html, tags=[], strip=True)
    body = Markup(body).unescape()
    if len(body) > 400:
        body = body[0:400] + "…"
    body = render_template_string(
        DOCUMENT_BODY_TEMPLATE, object_url=url_for(object), body=body, post=object
    )
    return Markup(body)



# coding=utf-8
""""""
from __future__ import absolute_import, division, print_function, \
    unicode_literals

from pathlib import Path

from abilian.core.models.subjects import User
from abilian.testing.util import login

from ..models import Document, Folder


def open_file(filename):
    path = Path(__file__).parent / "data" / "dummy_files" / filename
    return path.open("rb")


def test_document(app, session, req_ctx):
    root = Folder(title="root")
    doc = Document(parent=root, title="test")
    data = open_file("onepage.pdf").read()
    doc.set_content(data, "application/pdf")
    session.add(doc)
    session.commit()

    # coverage
    app.config["ANTIVIRUS_CHECK_REQUIRED"] = True
    doc.ensure_antivirus_scheduled()


def test_antivirus_properties(app, session, req_ctx):
    root = Folder(title="root")
    doc = Document(parent=root, title="test")
    doc.set_content(b"content", "text/plain")
    appcfg = app.config
    meta = doc.content_blob.meta

    # 1: not check required ################
    # no data in meta
    appcfg["ANTIVIRUS_CHECK_REQUIRED"] = False
    assert doc.antivirus_scanned is False
    assert doc.antivirus_status is None
    assert doc.antivirus_required is False
    assert doc.antivirus_ok is True

    # antivirus was run, but no result
    meta["antivirus"] = None
    assert doc.antivirus_scanned is True
    assert doc.antivirus_status is None
    assert doc.antivirus_required is False
    assert doc.antivirus_ok is True

    # virus detected
    meta["antivirus"] = False
    assert doc.antivirus_scanned is True
    assert doc.antivirus_status is False
    assert doc.antivirus_required is False
    assert doc.antivirus_ok is False

    # virus free
    meta["antivirus"] = True
    assert doc.antivirus_scanned is True
    assert doc.antivirus_status is True
    assert doc.antivirus_required is False
    assert doc.antivirus_ok is True

    # 2: check required ##################
    # no data in meta
    del meta["antivirus"]
    appcfg["ANTIVIRUS_CHECK_REQUIRED"] = True
    assert doc.antivirus_scanned is False
    assert doc.antivirus_status is None
    assert doc.antivirus_required is True
    assert doc.antivirus_ok is False

    # antivirus was run, but no result
    meta["antivirus"] = None
    assert doc.antivirus_scanned is True
    assert doc.antivirus_status is None
    assert doc.antivirus_required is True
    assert doc.antivirus_ok is False

    # virus detected
    meta["antivirus"] = False
    assert doc.antivirus_scanned is True
    assert doc.antivirus_status is False
    assert doc.antivirus_required is False
    assert doc.antivirus_ok is False

    # virus free
    meta["antivirus"] = True
    assert doc.antivirus_scanned is True
    assert doc.antivirus_status is True
    assert doc.antivirus_required is False
    assert doc.antivirus_ok is True


def test_folder_indexed(app, session, community1, community2):
    index_service = app.services["indexing"]
    index_service.start()

    security_service = app.services["security"]
    security_service.start()

    folder = Folder(title="Folder 1", parent=community1.folder)
    session.add(folder)

    folder_other = Folder(title="Folder 2: other", parent=community2.folder)
    session.add(folder_other)

    user_no_community = User(email="no_community@example.com", can_login=True)
    session.add(user_no_community)

    session.commit()

    svc = index_service
    obj_types = (Folder.entity_type,)

    with app.test_request_context():

        with login(user_no_community):
            res = svc.search("folder", object_types=obj_types)
            assert len(res) == 0

        with login(community1.test_user):
            res = svc.search("folder", object_types=obj_types)
            assert len(res) == 1
            hit = res[0]
            assert hit["object_key"] == folder.object_key
            assert hit["community_slug"] == community1.slug

        with login(community2.test_user):
            res = svc.search("folder", object_types=obj_types)
            assert len(res) == 1
            hit = res[0]
            assert hit["object_key"] == folder_other.object_key
            assert hit["community_slug"] == community2.slug



# coding=utf-8
"""
TODO: remove once the pytest-based tests all pass.
"""
from __future__ import absolute_import, print_function, unicode_literals

from abilian.services.security import Admin
from abilian.testing.util import path_from_url
from flask import url_for

from ..models import Community
from .base import CommunityIndexingTestCase as BaseIndexingTestCase


class CommunityWebTestCase(BaseIndexingTestCase):
    # FIXME later
    SQLALCHEMY_WARNINGS_AS_ERROR = False

    def test_index(self):
        with self.client_login(self.user.email, "azerty"):
            response = self.client.get(url_for("communities.index"))
            assert response.status_code == 200

    def test_community_home(self):
        url = self.app.default_view.url_for(self.community)
        user = self.user.email
        user_c2 = self.user_c2.email
        with self.client_login(user_c2, "azerty"):
            response = self.client.get(url)
            assert response.status_code == 403

        with self.client_login(user, "azerty"):
            response = self.client.get(url)
            assert response.status_code == 302
            expected_url = url_for(
                "wall.index", community_id=self.community.slug, _external=True
            )
            assert response.location == expected_url

    def test_community_settings(self):
        url = url_for("communities.settings", community_id=self.community.slug)
        with self.client_login(self.user.email, "azerty"):
            response = self.client.get(url)
            assert response.status_code == 403

            self.app.services["security"].grant_role(self.user, Admin)
            response = self.client.get(url)
            assert response.status_code == 200

            data = {
                "__action": "edit",
                "name": "edited community",
                "description": "my community",
                "linked_group": "",
                "type": "participative",
            }
            response = self.client.post(url, data=data)
            assert response.status_code == 302

            path = path_from_url(response.location)
            assert path == "/communities/" + self.community.slug + "/"

            community = Community.query.get(self.community.id)
            assert community.name == "edited community"

    def test_new(self):
        with self.client_login(self.user.email, "azerty"):
            response = self.client.get(url_for("communities.new"))
            assert response.status_code == 403

            self.app.services["security"].grant_role(self.user, Admin)
            response = self.client.get(url_for("communities.new"))
            assert response.status_code == 200

    def test_members(self):
        with self.client_login(self.user.email, "azerty"):
            url = url_for("communities.members", community_id=self.community.slug)
            response = self.client.get(url)
            assert response.status_code == 200

            # test add user
            data = {"action": "add-user-role", "user": self.user_c2.id}
            response = self.client.post(url, data=data)
            assert response.status_code == 403

            self.app.services["security"].grant_role(self.user, Admin)
            data = {
                "action": "add-user-role",
                "user": self.user_c2.id,
                "role": "member",
            }
            response = self.client.post(url, data=data)
            assert response.status_code == 302
            path = path_from_url(response.location)
            assert path == url

            membership = [
                m for m in self.community.memberships if m.user == self.user_c2
            ][0]
            assert membership.role == "member"

            data["action"] = "set-user-role"
            data["role"] = "manager"
            response = self.client.post(url, data=data)
            assert response.status_code == 302
            path = path_from_url(response.location)
            assert path == url

            self.session.expire(membership)
            assert membership.role == "manager"

            # Community.query.session is not self.db.session, but web app
            # session.
            community = Community.query.get(self.community.id)
            assert self.user_c2 in community.members

            # test delete
            data = {
                "action": "delete",
                "user": self.user_c2.id,
                "membership": [
                    m.id for m in community.memberships if m.user == self.user_c2
                ][0],
            }
            response = self.client.post(url, data=data)
            assert response.status_code == 302
            path = path_from_url(response.location)
            assert path == "/communities/" + self.community.slug + "/members"

            assert self.user_c2 not in community.members



# coding=utf-8
""""""
from __future__ import absolute_import, division, print_function, \
    unicode_literals

from datetime import datetime, timedelta
from unittest import TestCase

from abilian.core.models.subjects import User
from abilian.testing.util import client_login
from flask import url_for
from flask_login import login_user
from mock import Mock, patch
from pytest import mark
from six import text_type

from abilian.sbe.apps.communities.models import MANAGER, MEMBER
from abilian.sbe.apps.communities.tests.base import CommunityBaseTestCase
from abilian.sbe.apps.forum.tests.util import get_string_from_file

from ..commands import inject_email
from ..models import Post, Thread
from ..tasks import build_reply_email_address, extract_email_destination

pytest_plugins = ["abilian.sbe.apps.communities.tests.fixtures"]


def test_posts_ordering(db, community1):
    thread = Thread(community=community1, title="test ordering")
    db.session.add(thread)
    t1 = datetime(2014, 6, 20, 15, 0, 0)
    p1 = Post(thread=thread, body_html="post 1", created_at=t1)
    t2 = datetime(2014, 6, 20, 15, 1, 0)
    p2 = Post(thread=thread, body_html="post 2", created_at=t2)
    db.session.flush()
    p1_id, p2_id = p1.id, p2.id
    assert [p.id for p in thread.posts] == [p1_id, p2_id]

    # set post1 created after post2
    t1 = t1 + timedelta(minutes=2)
    p1.created_at = t1
    db.session.flush()
    db.session.expire(thread)  # force thread.posts refreshed from DB
    assert [p.id for p in thread.posts] == [p2_id, p1_id]


def test_thread_indexed(app, db, community1, community2, req_ctx):
    index_svc = app.services["indexing"]
    index_svc.start()
    security_svc = app.services["security"]
    security_svc.start()

    thread1 = Thread(title="Community 1", community=community1)
    db.session.add(thread1)

    thread2 = Thread(title="Community 2: other", community=community2)
    db.session.add(thread2)
    db.session.commit()

    index_svc = app.services["indexing"]
    obj_types = (Thread.entity_type,)

    login_user(community1.test_user)
    res = index_svc.search("community", object_types=obj_types)
    assert len(res) == 1
    hit = res[0]
    assert hit["object_key"] == thread1.object_key

    login_user(community2.test_user)
    res = index_svc.search("community", object_types=obj_types)
    assert len(res) == 1
    hit = res[0]
    assert hit["object_key"] == thread2.object_key


def test_forum_home(client, community1, login_admin, req_ctx):
    response = client.get(url_for("forum.index", community_id=community1.slug))
    assert response.status_code == 200


@mark.skip  # TODO: fixme later
def test_create_thread_informative(app, db, client, community1, req_ctx):
    """Test with 'informative' community.

    No mail sent, unless user is MANAGER
    """
    user = community1.test_user
    assert community1.type == "informative"
    community1.set_membership(user, MEMBER)
    db.session.commit()

    title = "Brand new thread"
    content = "shiny thread message"
    url = url_for("forum.new_thread", community_id=community1.slug)
    data = {"title": title, "message": content}
    data["__action"] = "create"

    mail = app.extensions["mail"]
    with client_login(client, user):
        with mail.record_messages() as outbox:
            data["send_by_email"] = "y"  # actually should not be in html form
            response = client.post(url, data=data)
            assert response.status_code == 302
            assert len(outbox) == 0

        community1.set_membership(user, MANAGER)
        db.session.commit()

        with mail.record_messages() as outbox:
            data["send_by_email"] = "y"  # should be in html form
            response = client.post(url, data=data)
            assert response.status_code == 302
            assert len(outbox) == 1

        with mail.record_messages() as outbox:
            del data["send_by_email"]
            response = client.post(url, data=data)
            assert response.status_code == 302
            assert len(outbox) == 0


def test_build_reply_email_address(app, req_ctx):
    app.config["MAIL_ADDRESS_TAG_CHAR"] = "+"

    post = Mock()
    post.id = 2
    post.thread_id = 3

    member = Mock()
    member.id = 4

    result = build_reply_email_address("test", post, member, "example.com")
    expected = "test+P-en-3-4-a8f33983311589176c711111dc38d94d@example.com"
    assert result == expected


def test_extract_mail_destination(app, req_ctx):
    app.config["MAIL_ADDRESS_TAG_CHAR"] = "+"
    # app.config['MAIL_SENDER'] = 'test@testcase.app.tld'

    test_address = "test+P-en-3-4-a8f33983311589176c711111dc38d94d@example.com"
    infos = extract_email_destination(test_address)
    assert infos == ("en", "3", "4")


class ViewTestCase(CommunityBaseTestCase):
    no_login = False
    SERVICES = ("security",)

    def test_create_thread_and_post(self):
        # activate email reply
        self.app.config["SBE_FORUM_REPLY_BY_MAIL"] = True
        self.app.config["MAIL_ADDRESS_TAG_CHAR"] = "+"

        # create a new user, add him/her to the current community as a MANAGER
        self.user = User(email="user_1@example.com", password="azerty", can_login=True)
        self.session.add(self.user)
        self.community.set_membership(self.user, MANAGER)
        self.session.commit()
        self.client_login(self.user.email, self.user.password)

        mail = self.app.extensions["mail"]
        with mail.record_messages() as outbox:
            title = "Brand new thread"
            content = "shiny thread message"
            url = url_for("forum.new_thread", community_id=self.community.slug)
            data = {"title": title, "message": content}
            data["__action"] = "create"
            data["send_by_email"] = "y"
            response = self.client.post(url, data=data)
            assert response.status_code == 302

            # extract the thread_id from the redirection url in response
            threadid = response.location.rsplit("/", 2)[1]

            # retrieve the new thread, make sur it has the message
            url = url_for(
                "forum.thread",
                thread_id=threadid,
                community_id=self.community.slug,
                title=title,
            )
            response = self.client.get(url)
            assert response.status_code == 200
            assert content in response.data.decode("utf-8")

            # check the email was sent with the new thread
            assert len(outbox) == 1
            assert outbox[0].subject == "[My Community] Brand new thread"

        # reset the outbox for checking threadpost email
        with mail.record_messages() as outbox:
            content = data["message"] = "my cherished post"
            del data["title"]
            response = self.client.post(url, data=data)
            assert response.status_code == 302

            # retrieve the new thread, make sur it has the message
            url = url_for(
                "forum.thread",
                thread_id=threadid,
                community_id=self.community.slug,
                title=title,
            )
            response = self.client.get(url)
            assert response.status_code == 200
            assert content in response.data.decode("utf-8")

            # check the email was sent with the new threadpost
            assert len(outbox) == 1
            expected = "[My Community] Brand new thread"
            assert text_type(outbox[0].subject) == expected


class CommandsTest(TestCase):
    @patch("fileinput.input")
    @patch("abilian.sbe.apps.forum.commands.process_email")
    def test_parse_forum_email(self, mock_process_email, mock_email):
        """No processing is tested only parsing into a email.message and
        verifying inject_email() logic."""
        # first load a test email returned by the mock_email
        mock_email.return_value = get_string_from_file("notification.email")

        # test the parsing function
        inject_email()

        # assert the email is read
        assert mock_email.called
        # assert a call on the celery task was made implying a message creation
        assert mock_process_email.delay.called

        ##
        mock_email.reset_mock()
        mock_process_email.delay.reset_mock()
        assert not mock_email.called
        assert not mock_process_email.delay.called

        mock_email.return_value = get_string_from_file("defects.email")
        inject_email()
        assert mock_email.called
        assert not mock_process_email.delay.called



# coding=utf-8
from __future__ import absolute_import, print_function, unicode_literals

from abilian.core.extensions import db
from flask import Blueprint, Response, make_response, render_template, request
from sqlalchemy.orm.exc import NoResultFound
from werkzeug.exceptions import NotFound, Unauthorized

from abilian.sbe.apps.documents.repository import repository

from .parser import Entry
from .renderer import Feed, to_xml

#
# Constants
#

ROOT = "http://localhost:5000/cmis/atompub"

# MIME TYPES (cf. section 3.3 of the CMIS specs)
MIME_TYPE_ATOM_FEED = "application/atom+xml;type=feed"
MIME_TYPE_ATOM_ENTRY = "application/atom+xml;type=entry"
MIME_TYPE_ATOM_SERVICE = "application/atomsvc+xml"
MIME_TYPE_CMIS_ATOM = "application/cmisatom+xml"
MIME_TYPE_CMIS_QUERY = "application/cmisquery+xml"
MIME_TYPE_CMIS_ALLOWABLE_ACTIONS = "application/cmisallowableactions+xml"
MIME_TYPE_CMIS_TREE = "application/cmistree+xml"
MIME_TYPE_CMIS_ACL = "application/cmisacl+xml"

atompub = Blueprint("cmis", __name__, url_prefix="/cmis/atompub")
route = atompub.route


#
# Dummy for now
#
class Logger(object):
    def debug(self, msg):
        print(msg)


log = Logger()


def log_result(result):
    print(78 * "-")
    print("Response:")
    print(result)
    print(78 * "-")


def produces(*mimetypes):
    def decorator(f):
        return f

    return decorator


def consumes(mimetype):
    def decorator(f):
        return f

    return decorator


# NOT WORKING
#  def decorator(f):
#    def g(*args, **kw):
#      assert mimetype == request.content_type
#      return f(*args, **kw)
#    return g
#  return decorator


# NOT WORKING
def render(template, status=200, mimetype=None):
    def decorator(f):
        def g(*args, **kw):
            print(f, args, kw)
            result = f(*args, **kw)
            rendered = render_template(template, **result)
            headers = {"Content-Type": mimetype}
            response = make_response(rendered, status, headers)
            print(response)
            return response

        return g

    return decorator


BOOLEAN_OPTIONS = ["includeAllowableActions", "includeACL", "includePolicyIds"]


def get_options(args):
    d = {}
    for k in BOOLEAN_OPTIONS:
        v = args.get(k, "false").strip()
        if v == "true":
            v = True
        elif v in ("false", ""):
            v = False
        else:
            raise Exception("Unexpected parameter value for %s: %s" % (k, v))
        d[k] = v
    return d


def get_document(id):
    doc = repository.get_document_by_id(id)
    if not doc:
        raise NotFound
    return doc


def get_folder(id):
    obj = repository.get_folder_by_id(id)
    if not obj:
        raise NotFound
    return obj


def get_object(id):
    obj = repository.get_object_by_id(id)
    if not obj:
        raise NotFound
    return obj


#
# Authentication (basic)
#
# @atompub.before_request
def authenticate():
    if not request.authorization:
        raise Unauthorized()

    username = request.authorization.username
    password = request.authorization.password

    if (username, password) != ("admin", "admin"):
        raise Unauthorized()


# @atompub.errorhandler(401)
def custom_401(error):
    print("custom_401")
    return Response(
        "Authentication required",
        401,
        {"WWWAuthenticate": 'Basic realm="Login Required"'},
    )


@atompub.errorhandler(NoResultFound)
def not_found_error_handler(error):
    """Converts SQLAlchemy NoResultFound exception to an HTTP error."""
    return "This object does not exist", 404


#
# Service Document
#
# @render("cmis/service.xml", mimetype=MIME_TYPE_ATOM_SERVICE)
@route("/")
@produces(MIME_TYPE_ATOM_SERVICE)
def getRepositoryInfo():
    log.debug("repositoryInfo called")

    root_folder = repository.root_folder
    ctx = {"ROOT": ROOT, "root_folder": root_folder}

    result = render_template("cmis/service.xml", **ctx)
    response = Response(result, mimetype=MIME_TYPE_ATOM_SERVICE)
    # return {'root': ROOT}
    return response


#
# Service Collections
#
@route("/types")
def getTypeChildren():
    log.debug("getTypeChildren called")

    result = render_template("cmis/types.xml", ROOT=ROOT)
    return Response(result, mimetype=MIME_TYPE_ATOM_FEED)


@route("/types", methods=["POST"])
def createType():
    log.debug("createType called")
    raise NotImplementedError()


#
# Entries (Documents, Folders, Relationships, Policies & Items)
#
@route("/entry")
def getObject():
    id = request.args.get("id")
    path = request.args.get("path")

    log.debug("getObject called on id=%s, path=%s" % (id, path))
    log.debug("URL: " + request.url)

    options = get_options(request.args)
    log.debug("Options: %s" % options)

    if id:
        obj = repository.get_object_by_id(id)
    elif path:
        obj = repository.get_object_by_path(path)
    else:
        raise NotFound("You must supply either an id or a path.")
    if not obj:
        raise NotFound("Object not found")

    result = to_xml(obj, **options)
    return Response(result, mimetype=MIME_TYPE_ATOM_ENTRY)


@route("/entry", methods=["PUT"])
def updateProperties():
    id = request.args.get("id")
    if not id:
        path = request.args.get("path")
    else:
        path = ""
    log.debug("updateProperties called on id=%s, path=%s" % (id, path))
    log.debug("URL: " + request.url)

    obj = get_object(id)
    return Response(to_xml(obj), mimetype=MIME_TYPE_ATOM_ENTRY)


@route("/entry", methods=["DELETE"])
def deleteObject():
    id = request.args.get("id")
    if not id:
        path = request.args.get("path")
    else:
        path = ""
    log.debug("deleteObject called on id=%s, path=%s" % (id, path))
    log.debug("URL: " + request.url)

    obj = get_object(id)

    # TODO: remove
    parent = obj.parent
    if parent:
        child_count_0 = len(parent.children)

    db.session.delete(obj)
    db.session.commit()

    if parent:
        child_count_1 = len(parent.children)
        assert child_count_1 == child_count_0 - 1

    return ("", 204, {})


#
# Content Streams Ressource
#
@route("/content")
def getContentStream():
    id = request.args.get("id")
    log.debug("getContentStream called on " + id)

    document = get_document(id)
    return Response(document.content, mimetype=document.content_type)


# setContentStream + appendContentStream
@route("/content", methods=["PUT"])
def setContentStream():
    id = request.args.get("id")
    log.debug("setContentStream called on " + id)

    document = get_document(id)
    created = document.content is None

    document.content = request.data
    document.content_type = request.content_type
    db.session.commit()

    if created:
        return ("", 201, {})
    else:
        return ("", 204, {})


@route("/content", methods=["DELETE"])
def deleteContentStream():
    id = request.args.get("id")
    log.debug("deleteContentStream called on " + id)

    document = get_document(id)
    document.content = None
    db.session.commit()

    return ("", 204, {})


#
# Allowable Actions Resource
#
@route("/allowableactions")
def getAllowableActions():
    id = request.args.get("id")
    log.debug("getAllowableActions called on " + id)

    obj = get_object(id)
    args = {"ROOT": ROOT, "object": obj}
    result = render_template("cmis/allowableactions.xml", **args)
    return Response(result, mimetype=MIME_TYPE_CMIS_ALLOWABLE_ACTIONS)


#
# Type Entries
#
@route("/type")
def getTypeDefinition():
    type_id = request.args.get("id")
    log.debug("getTypeDefinition called on " + type_id)

    if type_id == "cmis:document":
        result = render_template("cmis/type-document.xml", ROOT=ROOT)
    elif type_id == "cmis:folder":
        result = render_template("cmis/type-folder.xml", ROOT=ROOT)
    elif type_id == "cmis:relationship":
        result = render_template("cmis/type-relationship.xml", ROOT=ROOT)
    elif type_id == "cmis:policy":
        result = render_template("cmis/type-policy.xml", ROOT=ROOT)
    else:
        raise NotImplementedError()

    return Response(result, mimetype=MIME_TYPE_ATOM_ENTRY)


@route("/type", methods=["POST"])
def updateType():
    raise NotImplementedError()


@route("/type", methods=["DELETE"])
def deleteType():
    raise NotImplementedError()
    # return ("", 204, {})


#
# Folder Children collection
#
@route("/children")
def getChildren():
    id = request.args.get("id")
    log.debug("getChildren called on " + id)
    log.debug("URL: " + request.url)

    folder = get_folder(id)

    args = {"ROOT": ROOT, "folder": folder, "to_xml": to_xml}
    result = render_template("cmis/children.xml", **args)
    log_result(result)
    return Response(result, mimetype=MIME_TYPE_ATOM_ENTRY)


@route("/children", methods=["POST"])
def createObject():
    id = request.args.get("id")
    log.debug("createObject called on " + id)
    log.debug("URL: " + request.url)

    print("Received:")
    print(request.data)

    folder = get_folder(id)

    entry = Entry(request.data)
    name = entry.name
    type = entry.type

    if type == "cmis:folder":
        new_object = folder.create_subfolder(name)
        db.session.commit()

    elif type == "cmis:document":
        new_object = folder.create_document(name)
        if entry.content:
            new_object.content = entry.content
            new_object.content_type = entry.content_type
        db.session.commit()

    else:
        raise Exception("Unknown object type: %s" % type)

    result = to_xml(new_object)
    log_result(result)
    return Response(result, status=201, mimetype=MIME_TYPE_ATOM_ENTRY)
    # TODO:
    # URI newloc = null;
    # try {
    #   newloc = new URI(getBase() + "/node/" + newdoc.getId());
    # } catch (URISyntaxException e) {
    # // Shouldn't happen.
    # e.printStackTrace();
    # }

    # String output = getTemplate("entry.ftl").arg("entry", newdoc)
    # .arg("parent", parent).arg("objectType", objectType).render();

    # // XXX: .type() is here because of a bug in resteasy
    # return Response.created(newloc).entity(output).type(MIME_TYPE_ATOM_ENTRY).build();


#
# Feeds
#
@route("/parents")
def getObjectParents():
    """Object Parents Feed (GET)."""
    id = request.args.get("id")
    log.debug("getObjectParents called on " + id)

    obj = get_object(id)
    if obj.parent:
        feed = Feed(obj, [obj.parent])
    else:
        feed = Feed(obj, [])
    result = feed.to_xml()
    return Response(result, mimetype=MIME_TYPE_ATOM_FEED)


# Changes Feed (GET)
@route("/changes")
def getContentChanges():
    log.debug("getContentChanges called")
    raise NotImplementedError()


# Folder Descendants Feed (GET, DELETE)
@route("/descendants")
def getDescendants():
    id = request.args.get("id")
    log.debug("getObjectParents called on " + id)
    raise NotImplementedError()


@route("/descendants", methods=["DELETE"])
@route("/foldertree", methods=["DELETE"])
def deleteTree():
    id = request.args.get("id")
    log.debug("deleteTree called on " + id)

    obj = get_object(id)
    db.session.delete(obj)
    db.session.commit()

    return ("", 204, {})


# Folder Tree Feed (GET, DELETE: see above)
@route("/foldertree")
def getFolderTree():
    id = request.args.get("id")
    log.debug("getFolderTree called on " + id)
    raise NotImplementedError()


# All Versions Feed (GET)
def getAllVersions():
    pass


# Type Descendants Feed (GET)
@route("/typedesc")
def getTypeDescendants():
    type_id = request.args.get("typeId")
    log.debug("getTypeDescendants called on " + type_id)

    feed = Feed({}, [])
    result = feed.to_xml()
    return Response(result, mimetype=MIME_TYPE_ATOM_FEED)


#
# Query Collection
#
@route("/query", methods=["POST"])
def query():
    q = request.form.get("q")
    log.debug("query called: " + q)
    raise NotImplementedError()



# coding=utf-8
from __future__ import absolute_import, print_function, unicode_literals

from flask import url_for


def test_home(client, login_admin):
    response = client.get(url_for("social.home"))
    assert response.status_code == 200


def test_users(client, login_admin):
    response = client.get(url_for("social.users"))
    assert response.status_code == 200


def test_groups(client, login_admin):
    response = client.get(url_for("social.groups"))
    assert response.status_code == 200


def test_user(client, login_admin):
    user = login_admin
    response = client.get(url_for("social.user", user_id=user.id))
    assert response.status_code == 200



# coding=utf-8
"""Static configuration for the application.

TODO: add more (runtime) flexibility in plugin discovery, selection
and activation.
"""
from __future__ import absolute_import, division, print_function, \
    unicode_literals

import logging
import os
import subprocess
import sys
from pathlib import Path

import jinja2
from abilian.app import Application as BaseApplication
from abilian.core.celery import FlaskCelery as BaseCelery
from abilian.core.celery import FlaskLoader as CeleryBaseLoader
from abilian.core.commands import setup_abilian_commands
from abilian.core.extensions import db
from abilian.services import converter
from flask import current_app
from flask_script import Command, Manager
from werkzeug.serving import BaseWSGIServer

from .apps.documents.repository import repository
from .extension import sbe

# Used for side effects, do not remove

__all__ = ["create_app", "Application"]

logger = logging.getLogger(__name__)


def create_app(config=None):
    return Application(config=config)


command_manager = Manager(create_app)
setup_abilian_commands(command_manager)


# loader to be used by celery workers
class CeleryLoader(CeleryBaseLoader):
    flask_app_factory = "abilian.sbe.app.create_app"


celery = BaseCelery(loader=CeleryLoader)


class Application(BaseApplication):

    APP_PLUGINS = BaseApplication.APP_PLUGINS + (
        "abilian.sbe.apps.main",
        "abilian.sbe.apps.notifications",
        "abilian.sbe.apps.preferences",
        "abilian.sbe.apps.wiki",
        "abilian.sbe.apps.wall",
        "abilian.sbe.apps.documents",
        "abilian.sbe.apps.forum",
        # "abilian.sbe.apps.calendar",
        "abilian.sbe.apps.communities",
        "abilian.sbe.apps.social",
        "abilian.sbe.apps.preferences",
    )

    script_manager = command_manager

    def __init__(self, name="abilian_sbe", config=None, **kwargs):
        BaseApplication.__init__(self, name, config=config, **kwargs)
        loader = jinja2.PackageLoader("abilian.sbe", "templates")
        self.register_jinja_loaders(loader)

    def init_extensions(self):
        BaseApplication.init_extensions(self)
        sbe.init_app(self)
        repository.init_app(self)
        converter.init_app(self)


# SBE demo app bootstrap stuff
_SBE_DEMO_SCRIPT = """\
#!{BIN_DIR}/python
from __future__ import absolute_import

import sys
from abilian.sbe.app import command_entry_point

sys.exit(command_entry_point())
"""

_BASE_SERVER_ACTIVATE = BaseWSGIServer.server_activate


def _on_http_server_activate(self, *args, **kwargs):
    """This function is used as to monkey patch BaseWSGIServer.server_activate
    during `setup_sbe_demo`."""
    _BASE_SERVER_ACTIVATE(self)
    # now we are listening to socket
    host, port = self.server_address
    if host == "0.0.0.0":
        # chrome is not ok with 0.0.0.0
        host = "localhost"
    url = "http://{host}:{port}/setup".format(host=host, port=port)

    if sys.platform == "win32":
        os.startfile(url)
    else:
        opener = "open" if sys.platform == "darwin" else "xdg-open"
        subprocess.call([opener, url])


# run with python -m abilian.sbe.app setup_sbe_app
def setup_sbe_app():
    """Basic set up SBE application. Must be run inside a virtualenv.

    Will create `abilian_sbe` script, run a local server and open
    browser on app's setup wizard.
    """
    logger = logging.getLogger("sbe_demo")
    logger.setLevel(logging.INFO)

    if "VIRTUAL_ENV" not in os.environ:
        logger.error("Not in a virtualenv! Aborting.")
        return 1

    bin_dir = Path(sys.prefix) / "bin"

    if not bin_dir.exists() or not bin_dir.is_dir():
        logger.error("%s doesn't exists or is not a directory. Aborting", bin_dir)
        return 1

    script_file = bin_dir / "abilian_sbe"
    if script_file.exists():
        logger.info("%s already exists. Skipping creation.", script_file)
    else:
        with script_file.open("w") as out:
            logger.info('Create script: "%s".', script_file)
            content = _SBE_DEMO_SCRIPT.format(BIN_DIR=bin_dir)
            out.write(content)
        # 0755: -rwxr-xr-x
        script_file.chmod(0o755)

    current_app.config["PRODUCTION"] = True
    current_app.config["DEBUG"] = False
    current_app.config["ASSETS_DEBUG"] = False
    current_app.config["SITE_NAME"] = "Abilian SBE"
    current_app.config["MAIL_SENDER"] = "abilian-sbe-app@example.com"

    logger.info("Prepare CSS & JS files")
    command_manager.handle("abilian_sbe", ["assets", "build"])
    # disabled init config: only if not running setupwizard
    # command_manager.handle('abilian_sbe', ['config', 'init'])

    # patch server used to launch browser here immediately after socket opened
    BaseWSGIServer.server_activate = _on_http_server_activate
    return command_manager.handle("abilian_sbe", ["run", "--hide-config"])


def command_entry_point():
    command_manager.run(commands={"setup_sbe_app": Command(setup_sbe_app)})


if __name__ == "__main__":
    command_entry_point()



# coding=utf-8
"""Celery tasks related to document transformation and preview."""
from __future__ import absolute_import, division, print_function, \
    unicode_literals

import email
import mailbox
import re
from os.path import expanduser
from pathlib import Path
from typing import List, Text, Tuple

import bleach
import chardet
from abilian.core.celery import periodic_task
from abilian.core.extensions import db, mail
from abilian.core.models.subjects import User
from abilian.core.signals import activity
from abilian.core.util import md5, unwrap
from abilian.i18n import _l, render_template_i18n
from abilian.web import url_for
from celery import shared_task
from celery.schedules import crontab
from celery.utils.log import get_task_logger
from flask import current_app, g
from flask_babel import get_locale
from flask_mail import Message
from itsdangerous import Serializer
from six import text_type

from .forms import ALLOWED_ATTRIBUTES, ALLOWED_STYLES, ALLOWED_TAGS
from .models import Post, PostAttachment, Thread

MAIL_REPLY_MARKER = _l("_____Write above this line to post_____")

# logger = logging.getLogger(__package__)
# Celery logger
logger = get_task_logger(__name__)


def init_app(app):
    global check_maildir
    if app.config["INCOMING_MAIL_USE_MAILDIR"]:
        make_task = periodic_task(run_every=crontab(minute="*"))
        check_maildir = make_task(check_maildir)


@shared_task()
def send_post_by_email(post_id):
    """Send a post to community members by email."""
    with current_app.test_request_context("/send_post_by_email"):
        post = Post.query.get(post_id)
        if post is None:
            # deleted after task queued, but before task run
            return

        thread = post.thread
        community = thread.community
        logger.info(
            "Sending new post by email to members of community %r", community.name
        )

        CHUNK_SIZE = 20
        members_id = [member.id for member in community.members if member.can_login]
        chunk = []
        for idx, member_id in enumerate(members_id):
            chunk.append(member_id)
            if idx % CHUNK_SIZE == 0:
                batch_send_post_to_users.apply_async((post.id, chunk))
                chunk = []

        if chunk:
            batch_send_post_to_users.apply_async((post.id, chunk))


@shared_task(max_retries=10, rate_limit="12/m")
def batch_send_post_to_users(post_id, members_id, failed_ids=None):
    """Task run from send_post_by_email; auto-retry for mails that could not be
    successfully sent.

    Task default rate limit is 6/min.: there is at least 5 seconds between 2
    batches.

    During retry, if all `members_id` fails again, the task is retried
    5min. later, then 10, 20, 40... up to 10 times before giving up. This ensures
    retries up to approximatively 3 days and 13 hours after initial attempt
    (geometric series is: 5min * (1-2**10) / 1-2) = 5115 mins).
    """
    if not members_id:
        return

    post = Post.query.get(post_id)
    if post is None:
        # deleted after task queued, but before task run
        return

    failed = set()
    successfully_sent = []
    thread = post.thread
    community = thread.community
    user_filter = (
        User.id.in_(members_id) if len(members_id) > 1 else User.id == members_id[0]
    )
    users = User.query.filter(user_filter).all()

    for user in users:
        try:
            with current_app.test_request_context("/send_post_by_email"):
                send_post_to_user(community, post, user)
        except BaseException:
            failed.add(user.id)
        else:
            successfully_sent.append(user.id)

    if failed:
        if failed_ids is not None:
            failed_ids = set(failed_ids)

        if failed == failed_ids:
            # 5 minutes * (2** retry count)
            countdown = 300 * 2 ** batch_send_post_to_users.request.retries
            batch_send_post_to_users.retry([post_id, list(failed)], countdown=countdown)
        else:
            batch_send_post_to_users.apply_async([post_id, list(failed)])

    return {
        "post_id": post_id,
        "successfully_sent": successfully_sent,
        "failed": list(failed),
    }


def build_local_part(name, uid):
    """Build local part as 'name-uid-digest', ensuring length < 64."""
    tag = current_app.config["MAIL_ADDRESS_TAG_CHAR"]
    key = current_app.config["SECRET_KEY"]
    serializer = Serializer(key)
    signature = serializer.dumps(uid)
    digest = md5(signature)
    local_part = name + tag + uid + "-" + digest

    if len(local_part) > 64:
        if (len(local_part) - len(digest) - 1) > 64:
            # even without digest, it's too long
            raise ValueError(
                "Cannot build reply address: local part exceeds 64 " "characters"
            )
        local_part = local_part[:64]

    return local_part


def build_reply_email_address(name, post, member, domain):
    # type: (Text, Post, User, Text) -> Text
    """Build a reply-to email address embedding the locale, thread_id and
    user.id.

    :param name: (str)    first part of an email address
    :param post: Post()   to get post.thread_id
    :param member: User() to get user.id
    :param domain: (str)  the last domain name of the email address
    :return: (Unicode)    reply address for forum in the form
       test+P-fr-3-4-SDB7T5DXNZPD5YAHHVIKVOE2PM@testcase.app.tld

       'P' for 'post' - locale - thread id - user id - signature digest
    """
    locale = get_locale()
    uid = "-".join(["P", str(locale), str(post.thread_id), str(member.id)])
    local_part = build_local_part(name, uid)
    return local_part + "@" + domain


def extract_email_destination(address):
    # type: (Text) -> Tuple[Text, ...]
    """Return the values encoded in the email address.

    :param address: similar to test+IjEvMy8yLzQi.xjE04-4S0IzsdicTHKTAqcqa1fE@testcase.app.tld
    :return: List() of splitted values
    """
    local_part = address.rsplit("@", 1)[0]
    tag = current_app.config["MAIL_ADDRESS_TAG_CHAR"]
    name, ident = local_part.rsplit(tag, 1)
    uid, digest = ident.rsplit("-", 1)
    signed_local_part = build_local_part(name, uid)

    if local_part != signed_local_part:
        raise ValueError("Invalid signature in reply address")

    values = uid.split("-")
    header = values.pop(0)
    assert header == "P"
    return tuple(values)


def has_subtag(address):
    # type: (Text) -> bool
    """Return True if a subtag (defined in the config.py as
    'MAIL_ADDRESS_TAG_CHAR') was found in the name part of the address.

    :param address: email adress
    """
    name = address.rsplit("@", 1)[0]
    tag = current_app.config["MAIL_ADDRESS_TAG_CHAR"]
    return tag in name


def send_post_to_user(community, post, member):
    recipient = member.email
    subject = "[%s] %s" % (community.name, post.title)
    config = current_app.config
    sender = config.get("BULK_MAIL_SENDER", config["MAIL_SENDER"])
    SBE_FORUM_REPLY_BY_MAIL = config.get("SBE_FORUM_REPLY_BY_MAIL", False)
    SERVER_NAME = config.get("SERVER_NAME", "example.com")
    list_id = '"{} forum" <forum.{}.{}>'.format(
        community.name, community.slug, SERVER_NAME
    )
    forum_url = url_for("forum.index", community_id=community.slug, _external=True)
    forum_archive = url_for(
        "forum.archives", community_id=community.slug, _external=True
    )

    extra_headers = {
        "List-Id": list_id,
        "List-Archive": "<{}>".format(forum_archive),
        "List-Post": "<{}>".format(forum_url),
        "X-Auto-Response-Suppress": "All",
        "Auto-Submitted": "auto-generated",
    }

    if SBE_FORUM_REPLY_BY_MAIL and config["MAIL_ADDRESS_TAG_CHAR"] is not None:
        name = sender.rsplit("@", 1)[0]
        domain = sender.rsplit("@", 1)[1]
        replyto = build_reply_email_address(name, post, member, domain)
        msg = Message(
            subject,
            sender=sender,
            recipients=[recipient],
            reply_to=replyto,
            extra_headers=extra_headers,
        )
    else:
        msg = Message(
            subject, sender=sender, recipients=[recipient], extra_headers=extra_headers
        )

    ctx = {
        "community": community,
        "post": post,
        "member": member,
        "MAIL_REPLY_MARKER": MAIL_REPLY_MARKER,
        "SBE_FORUM_REPLY_BY_MAIL": SBE_FORUM_REPLY_BY_MAIL,
    }
    msg.body = render_template_i18n("forum/mail/new_message.txt", **ctx)
    msg.html = render_template_i18n("forum/mail/new_message.html", **ctx)
    logger.debug("Sending new post by email to %r", member.email)

    try:
        mail.send(msg)
    except BaseException:
        logger.error(
            "Send mail to user failed", exc_info=True
        )  # log to sentry if enabled


def extract_content(payload, marker):
    """Search the payload for marker, return content up to marker."""
    index = payload.rfind(marker)
    content = payload[:index]
    return content


def validate_html(payload):
    return bleach.clean(
        payload,
        tags=ALLOWED_TAGS,
        attributes=ALLOWED_ATTRIBUTES,
        styles=ALLOWED_STYLES,
        strip=True,
    ).strip()


def add_paragraph(newpost):
    """Add surrounding <p>newpost</p> if necessary."""
    newpost = newpost.strip()
    if not newpost.startswith("<p>"):
        newpost = "<p>" + newpost + "</p>"
    return newpost


def clean_html(newpost):
    # type: (Text) -> Text
    """Clean leftover empty blockquotes."""

    clean = re.sub(
        r"(<blockquote.*?<p>.*?</p>.*?</blockquote>)",
        "",
        newpost,
        flags=re.MULTILINE | re.DOTALL,
    )

    # this cleans auto generated reponse text (<br>timedate <a email /a><br>)
    # we reverse the string because re.sub replaces
    #  LEFTMOST NON-OVERLAPPING OCCURRENCES, and we only want the last match
    # in the string
    clean = re.sub(
        r"(>rb<.*?>a.*?=ferh\sa<.*?>rb<)",
        "",
        clean[::-1],
        flags=re.MULTILINE | re.DOTALL,
    )
    clean = clean[::-1]

    return clean


def decode_payload(part):
    # type: (email.message.Message) -> Text
    """Get the payload and decode (base64 & quoted printable)."""

    payload = part.get_payload(decode=True)

    if isinstance(payload, text_type):
        return payload

    # Please the typechecker (and make things a bit clearer)
    assert isinstance(payload, bytes)
    payload_bytes = payload  # type: bytes

    charset = part.get_content_charset()
    if charset is not None:
        try:
            payload_str = payload_bytes.decode(charset)
        except UnicodeDecodeError:
            payload_str = payload_bytes.decode("raw-unicode-escape")
    else:
        # What about other encodings? -> using chardet
        found = chardet.detect(payload_bytes)
        payload_str = payload_bytes.decode(found["encoding"])

    return payload_str


def process(message, marker):
    # type: (email.message.Message, Text) -> Tuple[Text, List[dict]]
    """Check the message for marker presence and return the text up to it if
    present.

    :raises LookupError otherwise.
    :return: sanitized html upto marker from message and attachements
    """
    assert isinstance(message, email.message.Message)
    content = {"plain": "", "html": ""}
    attachments = []
    # Iterate all message's parts for text/*
    for part in message.walk():
        content_type = part.get_content_type()
        content_disposition = part.get("Content-Disposition")

        if content_disposition is not None:
            attachments.append(
                {
                    "filename": part.get_filename(),
                    "content_type": part.get_content_type(),
                    "data": part.get_payload(decode=True),
                }
            )

        elif content_type in ["text/plain", "text/html"]:
            subtype = part.get_content_subtype()
            payload = decode_payload(part)
            content[subtype] += payload

    if marker in content["html"]:
        newpost = extract_content(content["html"], marker[:9])
        newpost = add_paragraph(validate_html(newpost))
        newpost = clean_html(newpost)
    elif marker in content["plain"]:
        newpost = extract_content(content["plain"], marker[:9])
        newpost = add_paragraph(newpost)
    else:
        raise LookupError("No marker:{} in email".format(marker))

    return newpost, attachments


@shared_task()
def process_email(message):
    # type: (email.message.Message) -> bool
    """Email.Message object from command line script Run message (parsed
    email).

    Processing chain extract community thread post member from reply_to
    persist post in db.
    """
    app = unwrap(current_app)
    # Extract post destination from To: field, (community/forum/thread/member)
    to_address = message["To"]

    assert isinstance(to_address, text_type)

    if not (has_subtag(to_address)):
        logger.info("Email %r has no subtag, skipping...", to_address)
        return False

    try:
        infos = extract_email_destination(to_address)
        locale = infos[0]
        thread_id = infos[1]
        user_id = infos[2]
    except BaseException:
        logger.error(
            "Recipient %r cannot be converted to locale/thread_id/user.id",
            to_address,
            exc_info=True,
        )
        return False

    # Translate marker with locale from email address
    rq_headers = [("Accept-Language", locale)]
    with app.test_request_context("/process_email", headers=rq_headers):
        marker = text_type(MAIL_REPLY_MARKER)

    # Extract text and attachments from message
    try:
        newpost, attachments = process(message, marker)
    except BaseException:
        logger.error("Could not Process message", exc_info=True)
        return False

    # Persist post
    with current_app.test_request_context("/process_email", headers=rq_headers):
        g.user = User.query.get(user_id)
        thread = Thread.query.get(thread_id)
        community = thread.community
        # FIXME: check membership, send back an informative email in case of an
        # error
        post = thread.create_post(body_html=newpost)
        obj_meta = post.meta.setdefault("abilian.sbe.forum", {})
        obj_meta["origin"] = "email"
        obj_meta["send_by_email"] = True
        activity.send(app, actor=g.user, verb="post", object=post, target=community)

        for desc in attachments:
            attachment = PostAttachment(name=desc["filename"])
            attachment.post = post
            attachment.set_content(desc["data"], desc["content_type"])
            db.session.add(attachment)
        db.session.commit()

    # Notify all parties involved
    send_post_by_email.delay(post.id)
    return True


def check_maildir():
    """Check the MailDir for emails to be injected in Threads.

    This task is registered only if `INCOMING_MAIL_USE_MAILDIR` is True.
    By default it is run every minute.
    """
    home = expanduser("~")
    maildirpath = str(Path(home) / "Maildir")
    src_mdir = mailbox.Maildir(maildirpath, factory=mailbox.MaildirMessage)

    src_mdir.lock()  # Useless but recommended if old mbox is used by error

    try:
        for key, message in src_mdir.items():
            processed = process_email(message)

            # delete the message if all went fine
            if processed:
                del src_mdir[key]

    finally:
        src_mdir.close()  # Flushes all changes to disk then unlocks



# coding=utf-8
"""
TODO: remove when refactoring done.
"""
from __future__ import absolute_import, division, print_function, \
    unicode_literals

from unittest import TestCase

from abilian.core.models.subjects import User
from flask import url_for
from mock import patch
from six import text_type

from abilian.sbe.apps.communities.models import MANAGER, MEMBER
from abilian.sbe.apps.communities.tests.base import CommunityBaseTestCase
from abilian.sbe.apps.forum.tests.util import get_string_from_file

from ..commands import inject_email


class NoLoginViewTest(CommunityBaseTestCase):
    """Test correct url response, without login or security involved."""

    def test(self):
        response = self.client.get(
            url_for("forum.index", community_id=self.community.slug)
        )
        assert response.status_code == 200


class ViewTestCase(CommunityBaseTestCase):
    no_login = False
    SERVICES = ("security",)

    def test_create_thread_and_post(self):
        # activate email reply
        self.app.config["SBE_FORUM_REPLY_BY_MAIL"] = True
        self.app.config["MAIL_ADDRESS_TAG_CHAR"] = "+"

        # create a new user, add him/her to the current community as a MANAGER
        self.user = User(email="user_1@example.com", password="azerty", can_login=True)
        self.session.add(self.user)
        self.community.set_membership(self.user, MANAGER)
        self.session.commit()
        self.client_login(self.user.email, self.user.password)

        mail = self.app.extensions["mail"]
        with mail.record_messages() as outbox:
            title = "Brand new thread"
            content = "shiny thread message"
            url = url_for("forum.new_thread", community_id=self.community.slug)
            data = {"title": title, "message": content}
            data["__action"] = "create"
            data["send_by_email"] = "y"
            response = self.client.post(url, data=data)
            assert response.status_code == 302

            # extract the thread_id from the redirection url in response
            threadid = response.location.rsplit("/", 2)[1]

            # retrieve the new thread, make sur it has the message
            url = url_for(
                "forum.thread",
                thread_id=threadid,
                community_id=self.community.slug,
                title=title,
            )
            response = self.client.get(url)
            assert response.status_code == 200
            assert content in response.data.decode("utf-8")

            # check the email was sent with the new thread
            assert len(outbox) == 1
            assert outbox[0].subject == "[My Community] Brand new thread"

        # reset the outbox for checking threadpost email
        with mail.record_messages() as outbox:
            content = data["message"] = "my cherished post"
            del data["title"]
            response = self.client.post(url, data=data)
            assert response.status_code == 302

            # retrieve the new thread, make sur it has the message
            url = url_for(
                "forum.thread",
                thread_id=threadid,
                community_id=self.community.slug,
                title=title,
            )
            response = self.client.get(url)
            assert response.status_code == 200
            assert content in response.data.decode("utf-8")

            # check the email was sent with the new threadpost
            assert len(outbox) == 1
            expected = "[My Community] Brand new thread"
            assert text_type(outbox[0].subject) == expected

    def test_create_thread_informative(self):
        """Test with 'informative' community.

        No mail sent, unless user is MANAGER
        """
        assert self.community.type == "informative"
        # create a new user, add him/her to the current community
        self.user = User(email="user_1@example.com", password="azerty", can_login=True)
        self.session.add(self.user)
        self.community.set_membership(self.user, MEMBER)
        self.session.commit()

        title = "Brand new thread"
        content = "shiny thread message"
        url = url_for("forum.new_thread", community_id=self.community.slug)
        data = {"title": title, "message": content}
        data["__action"] = "create"

        mail = self.app.extensions["mail"]
        self.client_login(self.user.email, self.user.password)

        with mail.record_messages() as outbox:
            data["send_by_email"] = "y"  # actually should not be in html form
            response = self.client.post(url, data=data)
            assert response.status_code == 302
            assert len(outbox) == 0

        self.community.set_membership(self.user, MANAGER)
        self.session.commit()

        with mail.record_messages() as outbox:
            data["send_by_email"] = "y"  # should be in html form
            response = self.client.post(url, data=data)
            assert response.status_code == 302
            assert len(outbox) == 1

        with mail.record_messages() as outbox:
            del data["send_by_email"]
            response = self.client.post(url, data=data)
            assert response.status_code == 302
            assert len(outbox) == 0

        self.client_logout()


class CommandsTest(TestCase):
    @patch("fileinput.input")
    @patch("abilian.sbe.apps.forum.commands.process_email")
    def test_parse_forum_email(self, mock_process_email, mock_email):
        """No processing is tested only parsing into a email.message and
        verifying inject_email() logic."""
        # first load a test email returned by the mock_email
        mock_email.return_value = get_string_from_file("notification.email")

        # test the parsing function
        inject_email()

        # assert the email is read
        assert mock_email.called
        # assert a call on the celery task was made implying a message creation
        assert mock_process_email.delay.called

        ##
        mock_email.reset_mock()
        mock_process_email.delay.reset_mock()
        assert not mock_email.called
        assert not mock_process_email.delay.called

        mock_email.return_value = get_string_from_file("defects.email")
        inject_email()
        assert mock_email.called
        assert not mock_process_email.delay.called



# coding=utf-8
"""
"""
from __future__ import absolute_import, print_function, unicode_literals

from distutils.command.build import build as _build

import setuptools
from pkg_resources import parse_requirements
from setuptools.command.develop import develop as _develop
from setuptools.command.sdist import sdist as _sdist

_install_requires = parse_requirements(open("requirements.in"))
install_requires = [str(req) for req in _install_requires]

_dev_requires = parse_requirements(open("etc/dev-requirements.txt"))
dev_requires = [str(req) for req in _dev_requires]

LONG_DESCRIPTION = open("README.rst", "r").read()


class build(_build):
    sub_commands = [("compile_catalog", None)] + _build.sub_commands


class sdist(_sdist):
    sub_commands = [("compile_catalog", None)] + _sdist.sub_commands


class develop(_develop):
    def run(self):
        _develop.run(self)
        self.run_command("compile_catalog")


setuptools.setup(
    name="abilian-sbe",
    use_scm_version=True,
    url="https://github.com/abilian/abilian-sbe",
    license="LGPL",
    author="Abilian SAS",
    author_email="contact@abilian.com",
    description="Social Business platform, including: document management, wiki, "
    "forum, enterprise social networking, and more",
    long_description=LONG_DESCRIPTION,
    packages=["abilian.sbe"],
    zip_safe=False,
    platforms="any",
    setup_requires=["babel", "setuptools-git", "setuptools_scm>=1.5.5"],
    install_requires=install_requires,
    extras_require={"tests": dev_requires, "dev": dev_requires},
    # dependency_links=dependency_links,
    include_package_data=True,
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Environment :: Web Environment",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: GNU Lesser General Public License v2 (LGPLv2)",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Framework :: Flask",
    ],
    cmdclass={"build": build, "sdist": sdist, "develop": develop},
)



# coding=utf-8
""""""
from __future__ import absolute_import, print_function, unicode_literals

import sqlalchemy as sa
from abilian.services.security import READ, Permission, security

from .models import CmisObject, Document, Folder


class SecurityException(Exception):
    pass


class Repository(object):
    """A simple document repository, implementing the basic functionalities of
    the CMIS model."""

    def __init__(self, app=None):
        if app is not None:
            self.init_app(app)

    def init_app(self, app):
        self.app = app
        app.extensions["content_repository"] = self

    @property
    def root_folder(self):
        folder = Folder.query.filter(Folder.parent == None).first()
        if folder:
            return folder

        # Should only happen during tests
        folder = Folder(title="root")
        return folder

    def get_object(self, id=None, path=None):
        """Get the CMIS object (document or folder) with either the given `id`
        or the given `path`.

        Returns None if the object doesn't exist.
        """
        if id:
            return self.get_object_by_id(id)
        else:
            return self.get_object_by_path(path)

    #
    # Id based navigation
    #
    def get_object_by_id(self, id):
        """Get the CMIS object (document or folder) with the given `id`.

        Returns None if the object doesn't exist.
        """
        obj = CmisObject.query.get(id)
        if obj is not None and not isinstance(obj, CmisObject):
            return None
        return obj

    def get_folder_by_id(self, id):
        """Get the folder with the given `id`.

        Returns None if the folder doesn't exist.
        """
        return Folder.query.get(id)

    def get_document_by_id(self, id):
        """Get the document with the given `id`.

        Returns None if the document doesn't exist.
        """
        document = Document.query.get(id)
        return document

    #
    # Path based navigation
    #
    def get_object_by_path(self, path):
        """Gets the CMIS object (document or folder) with the given `path`.

        Returns None if the object doesn't exist.
        """
        return self.root_folder.get_object_by_path(path)

    def get_folder_by_path(self, path):
        """Gets the folder with the given `path`.

        Returns None if the folder doesn't exist.
        """
        obj = self.root_folder.get_object_by_path(path)
        if obj is None or not obj.is_folder:
            return None
        else:
            return obj

    def get_document_by_path(self, path):
        """Gets the document with the given `path`.

        Returns None if the document doesn't exist.
        """
        obj = self.root_folder.get_object_by_path(path)
        if obj is None or not obj.is_document:
            return None
        else:
            return obj

    #
    # COPY / MOVE support
    #
    def copy_object(self, obj, dest_folder, dest_title=None):
        new_obj = obj.clone(title=dest_title, parent=dest_folder)
        if obj.is_folder:
            for child in obj.children:
                self.copy_object(child, new_obj)
        return new_obj

    def move_object(self, obj, dest_folder, dest_title=None):
        obj.parent = dest_folder
        if dest_title:
            obj.title = dest_title

    def rename_object(self, obj, title):
        obj.title = title

    def delete_object(self, obj):
        if obj.is_root_folder:
            raise Exception("Can't delete root folder.")

        session = sa.orm.object_session(obj)
        setattr(obj, "__path_before_delete", obj.path)  # for audit log.
        parent = obj.parent
        collection = parent.subfolders if obj.is_folder else parent.documents
        session.delete(obj)
        collection.remove(obj)

    #
    # Locking (TODO)
    #
    def is_locked(self, obj):
        return False

    def can_unlock(self, obj):
        return True

    def lock(self, obj):
        return "???"

    def unlock(self, obj):
        pass

    #
    # Security / access rights
    #
    def has_permission(self, user, permission, obj):
        assert isinstance(permission, Permission)
        return security.has_permission(user, permission, obj, inherit=True)

    def has_access(self, user, obj):
        """Checks that user has actual right to reach this object, 'read'
        permission on each of object's parents."""
        current = obj
        while current.parent is not None:
            if not self.has_permission(user, READ, current):
                return False
            current = current.parent
        return True


repository = Repository()



# coding=utf-8
from __future__ import absolute_import, print_function, unicode_literals

import re

from abilian.testing.util import client_login
from flask import g, url_for
from pytest import mark
from toolz import first

from abilian.sbe.apps.wiki import views

pytest_plugins = ["abilian.sbe.apps.communities.tests.fixtures"]


def test_home(client, community1, login_admin):
    response = client.get(url_for("wiki.index", community_id=community1.slug))
    assert response.status_code == 302

    response = client.get(
        url_for("wiki.page", title="Home", community_id=community1.slug)
    )
    assert response.status_code == 200


def test_create_page_initial_form(community1, user, req_ctx):
    g.user = user
    g.community = community1
    view = views.PageCreate()
    view.prepare_args([], {})
    form = view.form
    assert form["last_revision_id"].data is None


@mark.skip  # TODO: fixme later
def test_create_page(client, community1, req_ctx):
    community = community1
    user = community.test_user

    with client_login(client, user):
        title = "Some page name"
        url = url_for("wiki.page_new", community_id=community.slug)
        url += "?title=Some+page+name"
        response = client.get(url)
        assert response.status_code == 200
        # make sure the title is filled when coming from wikilink
        assert 'value="Some page name"' in response.get_data(as_text=True)

        title = "Some page"
        body = "LuuP3jai"
        url = url_for("wiki.page_new", community_id=community.slug)
        data = {"title": title, "body_src": body, "__action": "create"}
        response = client.post(url, data=data)
        assert response.status_code == 302

        url = url_for("wiki.page", community_id=community.slug, title=title)
        response = client.get(url)
        assert response.status_code == 200
        assert body in response.get_data(as_text=True)

        # edit
        url = url_for("wiki.page_edit", community_id=community.slug, title=title)
        response = client.get(url)
        assert response.status_code == 200

        # Slightly hackish way to get the page_id
        line = first(
            l
            for l in response.get_data(as_text=True).split("\n")
            if 'name="page_id"' in l
        )
        m = re.search('value="(.*?)"', line)
        page_id = int(m.group(1))

        url = url_for("wiki.page_changes", community_id=community.slug, title=title)
        response = client.get(url)
        assert response.status_code == 200

        url = url_for("wiki.page_source", community_id=community.slug, title=title)
        response = client.get(url)
        assert response.status_code == 200

        url = url_for("wiki.page_edit", community_id=community.slug)
        data = {"title": title, "page_id": page_id, "body_src": "abc def"}
        data["__action"] = "edit"
        response = client.post(url, data=data)
        assert response.status_code == 302

        url = url_for(
            "wiki.page_compare",
            rev0="on",
            rev1="on",
            community_id=community.slug,
            title=title,
        )
        response = client.get(url)
        assert response.status_code == 200



# coding=utf-8
""""""
from __future__ import absolute_import, print_function, unicode_literals

from abilian.services.security import Admin
from abilian.testing.util import client_login
from flask import url_for
from flask_login import current_user
from pytest import mark

from ..models import Community


def test_index(app, client, community1):
    security_service = app.services["security"]
    security_service.start()

    user = community1.test_user
    with client_login(client, user):
        response = client.get(url_for("communities.index"))
        assert response.status_code == 200


def test_community_home(app, client, community1, community2):
    security_service = app.services["security"]
    security_service.start()

    url = app.default_view.url_for(community1)

    user1 = community1.test_user
    with client_login(client, user1):
        response = client.get(url)
        assert response.status_code == 302
        expected_url = url_for(
            "wall.index", community_id=community1.slug, _external=True
        )
        assert response.location == expected_url

    user2 = community2.test_user
    with client_login(client, user2):
        response = client.get(url)
        assert response.status_code == 403


def test_new(app, client, community1):
    security_service = app.services["security"]
    security_service.start()

    user = community1.test_user

    with client_login(client, user):
        response = client.get(url_for("communities.new"))
        assert response.status_code == 403

        app.services["security"].grant_role(user, Admin)
        response = client.get(url_for("communities.new"))
        assert response.status_code == 200


@mark.skip  # TODO: fix
def test_community_settings(app, client, community1):
    security_service = app.services["security"]
    security_service.start()

    url = url_for("communities.settings", community_id=community1.slug)
    user = community1.test_user

    with client_login(client, user):
        response = client.get(url)
        assert response.status_code == 403

        app.services["security"].grant_role(user, Admin)
        response = client.get(url)
        assert response.status_code == 200

        data = {
            "__action": "edit",
            "name": "edited community",
            "description": "my community",
            "linked_group": "",
            "type": "participative",
        }
        response = client.post(url, data=data)
        assert response.status_code == 302
        expected_url = "http://localhost/communities/{}/".format(community1.slug)
        assert response.location == expected_url

        community = Community.query.get(community1.id)
        assert community.name == "edited community"


@mark.skip  # TODO: fix
def test_members(app, client, db, community1, community2):
    security_service = app.services["security"]
    security_service.start()

    user1 = community1.test_user
    user2 = community2.test_user

    with client_login(client, user1):
        url = url_for("communities.members", community_id=community1.slug)
        response = client.get(url)
        assert response.status_code == 200

        # test add user
        data = {"action": "add-user-role", "user": user2.id}
        response = client.post(url, data=data)
        assert response.status_code in (403, 400)

        app.services["security"].grant_role(user1, Admin)

        data = {"action": "add-user-role", "user": user2.id, "role": "member"}
        response = client.post(url, data=data)
        assert response.status_code == 302
        assert response.location == "http://localhost" + url

        membership = [m for m in community1.memberships if m.user == user2][0]
        assert membership.role == "member"

        data["action"] = "set-user-role"
        data["role"] = "manager"
        response = client.post(url, data=data)
        assert response.status_code == 302
        assert response.location == "http://localhost" + url

        db.session.expire(membership)
        assert membership.role == "manager"

        # Community.query.session is not self.db.session, but web app
        # session.
        community = Community.query.get(community1.id)
        assert user2 in community.members

        # test delete
        data = {
            "action": "delete",
            "user": user2.id,
            "membership": [m.id for m in community1.memberships if m.user == user2][0],
        }
        response = client.post(url, data=data)
        expected_url = "http://localhost/communities/{}/members".format(community1.slug)
        assert response.status_code == 302
        assert response.location == expected_url

        assert user2 not in community.members



# coding=utf-8
from __future__ import absolute_import, print_function, unicode_literals

DAV_PROPS = [
    "creationdate",
    "displayname",
    "getcontentlanguage",
    "getcontentlength",
    "getcontenttype",
    "getetag",
    "getlastmodified",
    "lockdiscovery",
    "resourcetype",
    "source",
    "supportedlock",
]

# DAV level 1
# OPTIONS = 'GET, HEAD, COPY, MOVE, POST, PUT, PROPFIND, PROPPATCH, OPTIONS, '\
#           'MKCOL, DELETE, TRACE, REPORT'
OPTIONS = (
    "GET, HEAD, POST, PUT, DELETE, OPTIONS, TRACE, PROPFIND, "
    + "PROPPATCH, MKCOL, COPY, MOVE"
)  # + ', LOCK, UNLOCK'

HTTP_CONTINUE = 100
HTTP_SWITCHING_PROTOCOLS = 101
HTTP_PROCESSING = 102

HTTP_OK = 200
HTTP_CREATED = 201
HTTP_ACCEPTED = 202
HTTP_NON_AUTHORITATIVE_INFO = 203
HTTP_NO_CONTENT = 204
HTTP_RESET_CONTENT = 205
HTTP_PARTIAL_CONTENT = 206
HTTP_MULTI_STATUS = 207
HTTP_IM_USED = 226

HTTP_MULTIPLE_CHOICES = 300
HTTP_MOVED = 301
HTTP_FOUND = 302
HTTP_SEE_OTHER = 303
HTTP_NOT_MODIFIED = 304
HTTP_USE_PROXY = 305
HTTP_TEMP_REDIRECT = 307
HTTP_BAD_REQUEST = 400
HTTP_PAYMENT_REQUIRED = 402
HTTP_FORBIDDEN = 403
HTTP_NOT_FOUND = 404
HTTP_METHOD_NOT_ALLOWED = 405
HTTP_NOT_ACCEPTABLE = 406
HTTP_PROXY_AUTH_REQUIRED = 407
HTTP_REQUEST_TIMEOUT = 408
HTTP_CONFLICT = 409
HTTP_GONE = 410
HTTP_LENGTH_REQUIRED = 411
HTTP_PRECONDITION_FAILED = 412
HTTP_REQUEST_ENTITY_TOO_LARGE = 413
HTTP_REQUEST_URI_TOO_LONG = 414
HTTP_MEDIATYPE_NOT_SUPPORTED = 415
HTTP_RANGE_NOT_SATISFIABLE = 416
HTTP_EXPECTATION_FAILED = 417
HTTP_UNPROCESSABLE_ENTITY = 422
HTTP_LOCKED = 423
HTTP_FAILED_DEPENDENCY = 424
HTTP_UPGRADE_REQUIRED = 426

HTTP_INTERNAL_ERROR = 500
HTTP_NOT_IMPLEMENTED = 501
HTTP_BAD_GATEWAY = 502
HTTP_SERVICE_UNAVAILABLE = 503
HTTP_GATEWAY_TIMEOUT = 504
HTTP_VERSION_NOT_SUPPORTED = 505
HTTP_INSUFFICIENT_STORAGE = 507
HTTP_NOT_EXTENDED = 510



# coding=utf-8
""""""
from __future__ import absolute_import, print_function, unicode_literals

from datetime import datetime

from flask import url_for
from pytest import mark

from abilian.sbe.apps.communities.tests.base import CommunityBaseTestCase, \
    CommunityIndexingTestCase

from ..models import Event


def test_create_event():
    start = datetime.now()
    event = Event(name="Test thread", start=start)
    assert event  # TODO


class IndexingTestCase(CommunityIndexingTestCase):
    @mark.skip
    def test_event_indexed(self):
        start = datetime.now()
        event1 = Event(name="Test event", community=self.community, start=start)
        event2 = Event(name="Test other event", community=self.c2, start=start)
        self.session.add(event1)
        self.session.add(event2)
        self.session.commit()

        svc = self.svc
        obj_types = (Event.entity_type,)
        with self.login(self.user_no_community):
            res = svc.search("event", object_types=obj_types)
            assert len(res) == 0

        with self.login(self.user):
            res = svc.search("event", object_types=obj_types)
            assert len(res) == 1
            hit = res[0]
            assert hit["object_key"] == event1.object_key

        with self.login(self.user_c2):
            res = svc.search("event", object_types=obj_types)
            assert len(res) == 1
            hit = res[0]
            assert hit["object_key"] == event2.object_key


class NoLoginViewTest(CommunityBaseTestCase):
    """Test correct url response, without login or security involved."""

    @mark.skip
    def test(self):
        response = self.client.get(
            url_for("calendar.index", community_id=self.community.slug)
        )
        assert response.status_code == 200
